# THIS FILE IS GENERATED FROM macrosynergy SETUP.PY
short_version: str = '0.0.34'
version: str = '0.0.34'
full_version: str = '0.0.34'
git_revision: str = '2e37c151e3bd7e30470e084de4507ab46980f45d'
release: bool = True
if not release:
    version = full_version



import numpy as np
import pandas as pd
from typing import List
from itertools import groupby
import random
from macrosynergy.management.shape_dfs import reduce_df
from macrosynergy.management.simulate_quantamental_data import make_qdf_black, make_qdf


def startend(dti, start, length):

    """Return start and end dates of a sequence as tuple

    :param <DateTimeIndex> dti: datetime series of working days
    :param <int> start: index of start
    :param <int> length: number of sequential days

    :return tuple of start and end date
    """

    tup = (dti[start], dti[start + (length - 1)])
    return tup


def make_blacklist(df: pd.DataFrame, xcat: str, cids: List[str] = None,
                   start: str = None, end: str = None,
                   nan_black: bool = False):

    """
    Converts binary category of standardized dataframe into a standardized dictionary
    that can serve as a blacklist for cross-sections in further analyses

    :param <pd.Dataframe> df: standardized DataFrame with following columns:
        'cid', 'xcat', 'real_date' and 'value'.
    :param <str> xcat: category with binary values, where 1 means blacklisting and 0
        means not blacklisting.
    :param List<str> cids: list of cross-sections that are considered in the formation
        of the blacklist. Per default, all available cross sections are considered.
    :param <str> start: earliest date in ISO format. Default is None and earliest date
        for which the respective category is available is used.
    :param <str> end: latest date in ISO format. Default is None and latest date
        for which the respective category is available is used.
    :param <bool> nan_black: if True NaNs are blacklisted (coverted to ones). Defaults is
        False, i.e. NaNs are converted to zeroes.

    :return <dict>: standardized dictionary with cross-sections as keys and tuples of
        start and end dates of the blacklist periods in ISO formats as values.
        If one cross section has multiple blacklist periods, numbers are added to the
        keys (i.e. TRY_1, TRY_2, etc.)
    """

    df["real_date"] = pd.to_datetime(df["real_date"], format="%Y-%m-%d")
    dfd = reduce_df(df=df, xcats=[xcat], cids=cids, start=start, end=end)
    assert all(np.isin(dfd.value.dropna().unique(), [0, 1])), \
        "blacklist values must all be 0/1"

    df_pivot = dfd.pivot(index='real_date', columns='cid', values='value')
    dates = df_pivot.index
    cids_df = list(df_pivot.columns)

    if nan_black:  # replace NaNs
        df_pivot = df_pivot.fillna(1)
    else:
        df_pivot = df_pivot.fillna(0)

    dates_dict = {}
    for cid in cids_df:
        count = 0
        column = df_pivot[cid].to_numpy()
        si = 0
        for k, v in groupby(column):  # iterator of consecutive keys and values
            v = list(v)  # instantiate the iterable in memory.
            length = len(v)
            if v[0] == 1:  # if blacklist period
                if count == 0:
                    dates_dict[cid] = startend(dates, si, length)
                elif count == 1:
                    val = dates_dict.pop(cid)
                    dates_dict[cid + '_1'] = val   # change key if more than 1 per cid
                    count += 1
                    dates_dict[cid + '_' + str(count)] = startend(dates, si, length)
                else:
                    dates_dict[cid + '_' + str(count)] = startend(dates, si, length)
                count += 1
            si += length
    return dates_dict


if __name__ == "__main__":

    cids = ['AUD', 'GBP', 'CAD', 'USD']
    cols = ['earliest', 'latest', 'mean_add', 'sd_mult']
    df_cid1 = pd.DataFrame(index=cids, columns=cols)

    df_cid1.loc['AUD'] = ['2010-01-01', '2020-12-31', 0, 1]
    df_cid1.loc['GBP'] = ['2011-01-01', '2020-11-30', 0, 1]
    df_cid1.loc['CAD'] = ['2011-01-01', '2021-11-30', 0, 1]
    df_cid1.loc['USD'] = ['2011-01-01', '2020-12-30', 0, 1]

    cols = ['earliest', 'latest', 'mean_add', 'sd_mult', 'ar_coef', 'back_coef']
    df_xcat1 = pd.DataFrame(index=['FXXR_NSA', 'FXCRY_NSA'],  columns=cols)
    df_xcat1.loc['FXXR_NSA'] = ['2010-01-01', '2020-12-31', 0, 1, 0, 0.3]
    df_xcat1.loc['FXCRY_NSA'] = ['2010-01-01', '2020-12-31', 0, 1, 0, 0.3]
    df1 = make_qdf(df_cid1, df_xcat1, back_ar=0.05)

    df_xcat2 = pd.DataFrame(index=['FXNONTRADE_NSA'],  columns=['earliest', 'latest'])
    df_xcat2.loc['FXNONTRADE_NSA'] = ['2010-01-01', '2021-11-30']
    black = {'AUD': ('2010-01-12', '2010-06-14'),
             'USD': ('2011-08-17', '2011-09-20'),
             'CAD_1': ('2011-01-04', '2011-01-23'),
             'CAD_2': ('2013-01-09', '2013-04-10'),
             'CAD_3': ('2015-01-12', '2015-03-12'),
             'CAD_4': ('2021-11-01', '2021-11-20')}

    print(black)
    df2 = make_qdf_black(df_cid1, df_xcat2, blackout=black)

    df = pd.concat([df1, df2]).reset_index()

    dates_dict = make_blacklist(df, xcat='FXNONTRADE_NSA', cids=None,
                                start=None, end=None)

    # If the output, from the below printed dictionary, differs from the above defined
    # dictionary, it should be by a date or two, as the construction of the dataframe,
    # using make_qdf_black(), will account for the dates received, in the dictionary,
    # being weekends. Therefore, if any of the dates, for the start or end of the
    # blackout period are Saturday or Sunday, the date for will be shifted to the
    # following Monday. Hence, a break in alignment from "blackout" to "dates_dict".
    print(dates_dict)



import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
from sklearn import metrics as skm
from scipy import stats
from typing import List, Union, Tuple
from datetime import timedelta
from collections import defaultdict
import warnings

from macrosynergy.management.simulate_quantamental_data import make_qdf
from macrosynergy.management.shape_dfs import reduce_df, categories_df


class SignalReturnRelations:

    """
    Class for analysing and visualizing signal and a return series.

    :param <pd.Dataframe> df: standardized DataFrame with the following necessary
        columns: 'cid', 'xcat', 'real_date' and 'value.
    :param <str> ret: return category.
    :param <str> sig: primary signal category for which detailed relational statistics
        can be calculated.
    :param <str, List[str]> rival_sigs: "rival signals" for which basic relational
        statistics can be calculated for comparison with the primary signal category. The
        table, if rival signals are defined, will be generated upon instantiation of the
        object.
        N.B.: parameters set for sig, such as sig_neg, freq, and agg_sig are equally
        applied to all rival signals.
    :param <bool> sig_neg: if set to True puts the signal in negative terms for all
        analysis. Default is False.
    :param <bool> cosp: If True the comparative statistics are calculated only for the
        "communal sample periods", i.e. periods and cross-sections that have values
        for all compared signals. Default is False.
    :param <str> start: earliest date in ISO format. Default is None in which case the
        earliest date available will be used.
    :param <str> end: latest date in ISO format. Default is None in which case the
        latest date in the df will be used.
    :param <dict> blacklist: cross-sections with date ranges that should be excluded from
        the data frame. If one cross-section has several blacklist periods append numbers
        to the cross-section code.
    :param <str> freq: letter denoting frequency at which the series are to be sampled.
        This must be one of 'D', 'W', 'M', 'Q', 'A'. Default is 'M'.
        The return series will always be summed over the sample period.
        The signal series will be aggregated according to the value of agg_sig.
    :param <str> agg_sig: aggregation method applied to the signal values in down-
        sampling. The default is "last".
        If defined, the additional signals will also use the same aggregation method for
        any down-sampling.
    :param <int> fwin: forward window of return category in base periods. Default is 1.
        This conceptually corresponds to the holding period of a position in
        accordance with the signal.
    :param <int> slip: implied slippage of feature availability for relationship with
        the target category. This mimics the relationship between trading signals and
        returns, which is often characterized by a delay due to the setup of of positions.
        Technically, this is a negative lag (early arrival) of the target category
        in working days prior to any frequency conversion. Default is 0.
  
    """
    def __init__(self, df: pd.DataFrame, ret: str, sig: str,
                 rival_sigs: Union[str, List[str]] = None, cids: List[str] = None,
                 sig_neg: bool = False, cosp: bool = False, start: str = None,
                 end: str = None, blacklist: dict = None, freq: str = 'M',
                  agg_sig: str = 'last', fwin: int = 1, slip: int = 0):

        if not isinstance(df, pd.DataFrame):
            raise TypeError(f"DataFrame expected and not {type(df)}.")
        
        df["real_date"] = pd.to_datetime(df["real_date"], format="%Y-%m-%d")

        self.dic_freq = {'D': 'daily', 'W': 'weekly', 'M': 'monthly',
                         'Q': 'quarterly', 'A': 'annual'}
        freq_error = f"Frequency parameter must be one of {list(self.dic_freq.keys())}."
        if not freq in self.dic_freq.keys():
            raise ValueError(freq_error)

        self.metrics = ['accuracy', 'bal_accuracy', 'pos_sigr', "pos_retr",
                        'pos_prec', 'neg_prec', 'pearson', 'pearson_pval',
                        'kendall', 'kendall_pval']

        self.ret = ret
        self.freq = freq

        assert isinstance(cosp, bool), f"<bool> object expected and not {type(cosp)}."
        self.cosp = cosp
        self.start = start
        self.end = end
        self.blacklist = blacklist
        self.fwin = fwin

        xcats = list(df['xcat'].unique())
        assert sig in xcats, "Primary signal must be available in the DataFrame."
        self.sig = sig

        signals = [self.sig]
        if rival_sigs is not None:

            r_sigs_error = "Signal or list of signals expected."
            assert isinstance(rival_sigs, (str, list)), r_sigs_error

            r_sigs = [rival_sigs] if isinstance(rival_sigs, str) else rival_sigs

            intersection = set(xcats).intersection(r_sigs)
            missing = set(r_sigs).difference(intersection)

            rival_error = f"The additional signals must be present in the defined " \
                          f"DataFrame. It is currently missing, {missing}."
            assert set(r_sigs).issubset(set(xcats)), rival_error
            signals += r_sigs

        self.signals = signals

        xcats = self.signals + [ret]

        dfd = reduce_df(
            df, xcats=xcats, cids=cids, start=start, end=end, blacklist=blacklist
        )
        
        # Since there may be any metrics in the DF at this point, simply apply slip to all.
        metric_cols: List[str] = list(set(dfd.columns.tolist()) 
                                  - set(['real_date', 'xcat', 'cid']))
        dfd: pd.DataFrame = self.apply_slip(
            target_df=dfd, slip=slip,
            cids=cids, xcats=xcats,
            metrics=metric_cols
        )

        # Naturally, only applicable if rival signals have been passed.
        if self.cosp and len(signals) > 1:
            dfd = self.__communal_sample__(df=dfd)

        self.dfd = dfd

        df = categories_df(
            dfd, xcats=xcats, cids=cids, val='value', start=None, end=None,
            freq=freq, blacklist=None, lag=1, fwin=fwin, xcat_aggs=[agg_sig, 'sum']
        )
        self.df = df
        self.cids = list(np.sort(self.df.index.get_level_values(0).unique()))

        if sig_neg:
            self.df.loc[:, self.signals] *= -1
            s_copy = self.signals.copy()

            self.signals = [s + "_NEG" for s in self.signals]
            self.sig += "_NEG"
            self.df.rename(columns=dict(zip(s_copy, self.signals)), inplace=True)

        if len(self.signals) > 1:

            self.df_sigs = self.__rival_sigs__()

        self.df_cs = self.__output_table__(cs_type='cids')
        self.df_ys = self.__output_table__(cs_type='years')

    @staticmethod
    def __slice_df__(df: pd.DataFrame, cs: str, cs_type: str):
        """
        Slice DataFrame by year, cross-section, or use full panel.

        :param <pd.DataFrame> df: standardised DataFrame.
        :param <str> cs: individual segment, cross-section or year.
        :param <str> cs_type: segmentation type.
        """

        # Row names of cross-sections or years.
        if cs != 'Panel' and cs_type == 'cids':
            df_cs = df.loc[cs]
        elif cs != 'Panel':
            df_cs = df[df['year'] == float(cs)]
        else:
            df_cs = df

        return df_cs
    
    @classmethod
    def apply_slip(self, target_df: pd.DataFrame, slip: int,
                    cids: List[str], xcats: List[str],
                    metrics: List[str]) -> pd.DataFrame:
        """
        Applied a slip, i.e. a negative lag, to the target DataFrame 
        for the given cross-sections and categories, on the given metrics.
        
        Parameters
        :param <pd.DataFrame> target_df: DataFrame to which the slip is applied.
        :param <int> slip: Slip to be applied.
        :param <List[str]> cids: List of cross-sections.
        :param <List[str]> xcats: List of categories.
        :param <List[str]> metrics: List of metrics to which the slip is applied.
        :return <pd.DataFrame> target_df: DataFrame with the slip applied.
        :raises <TypeError>: If the provided parameters are not of the expected type.
        :raises <ValueError>: If the provided parameters are semantically incorrect.
        """

        target_df = target_df.copy(deep=True)
        if not (isinstance(slip, int) and slip >= 0):
            raise ValueError("Slip must be a non-negative integer.")
        
        if cids is None:
            cids = target_df['cid'].unique().tolist()
        if xcats is None:
            xcats = target_df['xcat'].unique().tolist()

        sel_tickers : List[str] = [f"{cid}_{xcat}" for cid in cids for xcat in xcats]
        target_df['tickers'] = target_df['cid'] + '_' + target_df['xcat']

        if not set(sel_tickers).issubset(set(target_df['tickers'].unique())):
            warnings.warn("Tickers targetted for applying slip are not present in the DataFrame.\n"
             f"Missing tickers: {sorted(list(set(sel_tickers) - set(target_df['tickers'].unique())))}")

        slip : int = slip.__neg__()
        
        target_df[metrics] = target_df.groupby('tickers')[metrics].shift(slip)
        target_df = target_df.drop(columns=['tickers'])
        
        return target_df

    def __table_stats__(self, df_segment: pd.DataFrame, df_out: pd.DataFrame,
                        segment: str, signal: str):
        """
        Method used to compute the evaluation metrics across segments: cross-section,
        yearly or category level.

        :param <pd.DataFrame> df_segment: segmented DataFrame.
        :param <pd.DataFrame> df_out: metric DataFrame where the index will be all
            segments for the respective segmentation type.
        :param <str> segment: segment which could either be an individual cross-section,
            year or category. Will form the index of the returned DataFrame.
        :param <str> signal: signal category.
        """

        # Account for NaN values between the single respective signal and return. Only
        # applicable for rival signals panel level calculations.
        df_segment = df_segment.loc[:, [self.ret, signal]].dropna(axis=0, how="any")

        df_sgs = np.sign(df_segment.loc[:, [self.ret, signal]])
        # Exact zeroes are disqualified for sign analysis only.
        df_sgs = df_sgs[~((df_sgs.iloc[:, 0] == 0) | (df_sgs.iloc[:, 1] == 0))]

        sig_sign = df_sgs[signal]
        ret_sign = df_sgs[self.ret]

        df_out.loc[segment, "accuracy"] = skm.accuracy_score(sig_sign, ret_sign)
        df_out.loc[segment, "bal_accuracy"] = skm.balanced_accuracy_score(
            sig_sign, ret_sign
        )

        df_out.loc[segment, "pos_sigr"] = np.mean(sig_sign == 1)
        df_out.loc[segment, "pos_retr"] = np.mean(ret_sign == 1)
        df_out.loc[segment, "pos_prec"] = skm.precision_score(
            ret_sign, sig_sign, pos_label=1
        )
        df_out.loc[segment, "neg_prec"] = skm.precision_score(
            ret_sign, sig_sign, pos_label=-1
        )

        ret_vals, sig_vals = df_segment[self.ret], df_segment[signal]
        df_out.loc[segment, ["kendall", "kendall_pval"]] = stats.kendalltau(ret_vals,
                                                                            sig_vals)
        corr, corr_pval = stats.pearsonr(ret_vals, sig_vals)
        df_out.loc[segment, ["pearson", "pearson_pval"]] = np.array([corr, corr_pval])

        return df_out

    def __communal_sample__(self, df: pd.DataFrame):
        """
        On a multi-index DataFrame, where the outer index are the cross-sections and the
        inner index are the timestamps, exclude any row where all signals do not have
        a realised value.

        :param <pd.Dataframe> df: standardized DataFrame with the following necessary
            columns: 'cid', 'xcat', 'real_date' and 'value'.

        NB.:
        Remove the return category from establishing the intersection to preserve the
        maximum amount of signal data available (required because of the applied lag).
        """

        df_w = df.pivot(index=('cid', 'real_date'), columns='xcat', values="value")

        storage = []
        for c, cid_df in df_w.groupby(level=0):
            cid_df = cid_df[self.signals + [self.ret]]

            final_df = pd.DataFrame(
                data=np.empty(shape=cid_df.shape), columns=cid_df.columns,
                index=cid_df.index
            )
            final_df.loc[:, :] = np.NaN

            # Return category is preserved.
            final_df.loc[:, self.ret] = cid_df[self.ret]

            intersection_df = cid_df.loc[:, self.signals].droplevel(level=0)
            # Intersection exclusively across the signals.
            intersection_df = intersection_df.dropna(how="any")
            s_date = intersection_df.index[0]
            e_date = intersection_df.index[-1]

            final_df.loc[(c, s_date): (c, e_date), self.signals] = intersection_df.to_numpy()
            storage.append(final_df)

        df = pd.concat(storage)
        df = df.stack().reset_index().sort_values(["cid", "xcat", "real_date"])
        df.columns = ["cid", "real_date", "xcat", "value"]

        return df[["cid", "xcat", "real_date", "value"]]

    def __output_table__(self, cs_type: str = 'cids'):
        """
        Creates a DataFrame with information on the signal-return relation across
        cross-sections or years and, additionally, the panel.

        :param <str> cs_type: the segmentation type.

        """

        # Analysis completed exclusively on the primary signal.
        df = self.df[[self.ret, self.sig]]

        # Will remove any timestamps where both the signal & return are not realised.
        # Applicable even if communal sampling has been applied given the alignment
        # excludes the return category.
        df = df.dropna(how="any")

        if cs_type == "cids":
            css = self.cids
        else:
            df["year"] = np.array(df.reset_index(level=1)["real_date"].dt.year)
            css = [str(y) for y in list(set(df["year"]))]
            css = sorted(css)

        statms = self.metrics
        df_out = pd.DataFrame(index=["Panel", "Mean", "PosRatio"] + css, columns=statms)

        for cs in (css + ["Panel"]):

            df_cs = self.__slice_df__(df=df, cs=cs, cs_type=cs_type)
            df_out = self.__table_stats__(df_segment=df_cs, df_out=df_out, segment=cs,
                                          signal=self.sig)

        df_out.loc["Mean", :] = df_out.loc[css, :].mean()

        above50s = statms[0:6]
        # Overview of the cross-sectional performance.
        df_out.loc["PosRatio", above50s] = (df_out.loc[css, above50s] > 0.5).mean()

        above0s = statms[6::2]
        pos_corr_coefs = df_out.loc[css, above0s] > 0
        df_out.loc["PosRatio", above0s] = pos_corr_coefs.mean()

        below50s = statms[7::2]
        pvals_bool = df_out.loc[css, below50s] < 0.5
        pos_pvals = np.mean(np.array(pvals_bool) * np.array(pos_corr_coefs), axis=0)
        # Positive correlation with error prob < 50%.
        df_out.loc["PosRatio", below50s] = pos_pvals

        return df_out.astype("float")

    def __rival_sigs__(self):
        """
        Produces the panel-level table for the additional signals.
        """

        df_out = pd.DataFrame(index=self.signals, columns=self.metrics)
        df = self.df

        for s in self.signals:
            # Entire panel will be passed in.
            df_out = self.__table_stats__(
                df_segment=df, df_out=df_out, segment=s, signal=s
            )

        return df_out

    def signals_table(self, sigs: List[str] = None):
        """
        Output table on relations of various signals with the target return.

        :param <List[str]> sigs: signal categories to be included in the panel-level
            table. Default is None and all present signals will be displayed. Alternative
            is a valid subset of the possible categories. Primary signal must be passed
            if to be included.

        NB.:
        Analysis will be based exclusively on the panel level. Will only return a table
        if rival signals have been defined upon instantiation. If the communal sample
        parameter has been set to True, all signals will be aligned on the individual
        cross-sectional level.
        """

        try:
            df_sigs = self.df_sigs.round(decimals=3)
        except Exception:
            error_msg = "Additional signals have not been defined on the instance."
            raise AttributeError(error_msg)
        else:
            # Set to all available signals.
            sigs = self.signals if sigs is None else sigs

            assert isinstance(sigs, list), "List of signals expected."

            sigs_error = f"The requested signals must be a subset of the primary plus " \
                         f"additional signals received, {self.signals}."
            assert set(sigs).issubset(set(self.signals)), sigs_error

            return df_sigs.loc[sigs, :]

    def cross_section_table(self):
        """ Output table on relations across sections and the panel. """

        return self.df_cs.round(decimals=3)

    def yearly_table(self):
        """ Output table on relations across sections and the panel. """
        return self.df_ys.round(decimals=3)

    @staticmethod
    def __yaxis_lim__(accuracy_df: pd.DataFrame):
        """Determines the range the y-axis is defined over.

        :param <pd.DataFrame> accuracy_df: two dimensional DataFrame with accuracy &
            balanced accuracy columns.

        N.B.: The returned range will always be below 0.5.
        """
        y_axis = lambda min_correl: min_correl > 0.45
        min_value = accuracy_df.min().min()
        # Ensures any accuracy statistics greater than 0.5 are more pronounced given the
        # adjusted scale.
        y_input = 0.45 if y_axis(min_value) else min_value

        return y_input

    def accuracy_bars(self, type: str = "cross_section", title: str = None,
                      size: Tuple[float] = None,
                      legend_pos: str = "best"):
        """
        Plot bar chart for the overall and balanced accuracy metrics.

        :param <str> type: type of segment over which bars are drawn. Either
            "cross_section" (default), "years" or "signals".
        :param <str> title: chart header - default will be applied if none is chosen.
        :param <Tuple[float]> size: 2-tuple of width and height of plot - default will be
            applied if none is chosen.
        :param <str> legend_pos: position of legend box. Default is 'best'.
            See the documentation of matplotlib.pyplot.legend.

        """

        assert type in ["cross_section", "years", "signals"]

        if type == "cross_section":
            df_xs = self.df_cs
        elif type == "years":
            df_xs = self.df_ys
        else:
            df_xs = self.df_sigs

        dfx = df_xs[~df_xs.index.isin(['PosRatio'])]

        if title is None:
            refsig = "various signals" if type == "signals" else self.sig
            title = f"Accuracy for sign prediction of {self.ret} based on {refsig} " \
                    f"at {self.dic_freq[self.freq]} frequency."
        if size is None:
            size = (np.max([dfx.shape[0] / 2, 8]), 6)

        plt.style.use("seaborn")
        plt.figure(figsize=size)
        x_indexes = np.arange(dfx.shape[0])

        w = 0.4
        plt.bar(
            x_indexes - w / 2, dfx['accuracy'], label='Accuracy', width=w,
            color='lightblue'
        )
        plt.bar(
            x_indexes + w / 2, dfx['bal_accuracy'], label='Balanced Accuracy', width=w,
            color='steelblue'
        )

        plt.xticks(ticks=x_indexes, labels=dfx.index, rotation=0)
        plt.axhline(y=0.5, color='black', linestyle='-', linewidth=0.5)

        y_input = self.__yaxis_lim__(
            accuracy_df=dfx.loc[:, ['accuracy', 'bal_accuracy']]
        )

        plt.ylim(round(y_input, 2))

        plt.title(title)
        plt.legend(loc=legend_pos)
        plt.show()

    def correlation_bars(self, type: str = "cross_section", title: str = None,
                         size: Tuple[float] = None,
                         legend_pos: str = "best"):
        """
        Plot correlation coefficients and significance.

        :param <str> type: type of segment over which bars are drawn. Either
            "cross_section" (default), "years" or "signals".
        :param <str> title: chart header. Default will be applied if none is chosen.
        :param <Tuple[float]> size: 2-tuple of width and height of plot.
            Default will be applied if none is chosen.
        :param <str> legend_pos: position of legend box. Default is 'best'.
            See matplotlib.pyplot.legend.

        """
        assert type in ["cross_section", "years", "signals"]

        if type == "cross_section":
            df_xs = self.df_cs
        elif type == "years":
            df_xs = self.df_ys
        else:
            df_xs = self.df_sigs

        # Panel plus the cs_types.
        dfx = df_xs[~df_xs.index.isin(['PosRatio', 'Mean'])]

        pprobs = np.array(
            [(1 - pv) * (np.sign(cc) + 1) / 2 for pv, cc in zip(dfx["pearson_pval"],
                                                                dfx["pearson"])]
        )
        pprobs[pprobs == 0] = 0.01
        kprobs = np.array(
            [(1 - pv) * (np.sign(cc) + 1) / 2 for pv, cc in zip(dfx["kendall_pval"],
                                                                dfx["kendall"])]
        )

        kprobs[kprobs == 0] = 0.01

        if title is None:
            refsig = "various signals" if type == "signals" else self.sig
            title = f"Positive correlation probability of {self.ret} " \
                    f"and lagged {refsig} at {self.dic_freq[self.freq]} frequency."
        if size is None:
            size = (np.max([dfx.shape[0]/2, 8]), 6)

        plt.style.use('seaborn')
        plt.figure(figsize=size)
        x_indexes = np.arange(len(dfx.index))
        w = 0.4
        plt.bar(x_indexes - w / 2, pprobs, label='Pearson', width=w, color='lightblue')
        plt.bar(x_indexes + w / 2, kprobs, label='Kendall', width=w, color='steelblue')
        plt.xticks(ticks=x_indexes, labels=dfx.index, rotation=0)

        plt.axhline(
            y=0.95, color='orange', linestyle='--', linewidth=0.5,
            label='95% probability'
        )
        plt.axhline(
            y=0.99, color='red', linestyle='--', linewidth=0.5, label='99% probability'
        )

        plt.title(title)
        plt.legend(loc=legend_pos)
        plt.show()

    def summary_table(self):
        """
        Return summary output table of signal-return relations.

        N.B.: The interpretation of the columns is generally as follows:

        accuracy refers to accuracy for binary classification, i.e. positive or negative
            return, and gives the ratio of correct prediction of the sign of returns
            against all predictions. Note that exact zero values for either signal or
            return series will not be considered for accuracy analysis.
        bal_accuracy refers to balanced accuracy. This is the average of the ratios of
            correctly detected positive returns and correctly detected negative returns.
            The denominators here are the total of actual positive and negative returns
            cases. Technically, this is the average of sensitivity and specificity.
        pos_sigr is the ratio of positive signals to all predictions. It indicates the
            long bias of the signal.
        pos_retr is the ratio of positive returns to all observed returns. It indicates
            the positive bias of the returns.
        pos_prec means positive precision, i.e. the ratio of correct positive return
            predictions to all positive predictions. It indicates how well the positive
            predictions of the signal have fared. Generally, good positive precision is
            easy to accomplish if the ratio of positive returns has been high (the signal
            can simply propose a long position for the full duration).
        neg_prec means negative precision, i.e. the ratio of correct negative return
            predictions to all negative predictions. It indicates how well the negative
            predictions of the signal have fared. Generally, good negative precision is
            hard to accomplish if the ratio of positive returns has been high.
        pearson is the Pearson correlation coefficient between signal and subsequent
            return.
        pearson_pval is the probability that the (positive) correlation has been
            accidental, assuming that returns are independently distributed. This
            statistic would be invalid for forward moving averages.
        kendall is the Kendall correlation coefficient between signal and subsequent
            return.
        kendall_pval is the probability that the (positive) correlation has been
            accidental, assuming that returns are independently distributed. This
            statistic would be invalid for forward moving averages.

        The rows have the following meaning:

        Panel refers to the the whole panel of cross-sections and sample period,
            excluding unavailable and blacklisted periods.
        Mean years is the mean of the statistic across all years.
        Mean cids is the mean of the statistic across all cross-sections.
        Positive ratio is the ratio of positive years or cross-sections for which the
            statistic was above its "neutral" level, i.e. above 0.5 for classification
            ratios and positive correlation probabilities, and above 0 for the
            correlation coefficients.
        """

        dfys = self.df_ys.round(decimals=5)
        dfcs = self.df_cs.round(decimals=5)
        dfsum = pd.concat([dfys.iloc[:3, ], dfcs.iloc[1:3, ]], axis=0)

        dfsum.index = ["Panel", "Mean years", "Positive ratio",
                       "Mean cids", "Positive ratio"]

        return dfsum


if __name__ == "__main__":

    cids = ['AUD', 'CAD', 'GBP', 'NZD']
    xcats = ['XR', 'CRY', 'GROWTH', 'INFL']
    df_cids = pd.DataFrame(index=cids,
                           columns=['earliest', 'latest', 'mean_add', 'sd_mult'])
    df_cids.loc['AUD'] = ['2000-01-01', '2020-12-31', 0, 1]
    df_cids.loc['CAD'] = ['2001-01-01', '2020-11-30', 0, 1]
    df_cids.loc['GBP'] = ['2002-01-01', '2020-11-30', 0, 2]
    df_cids.loc['NZD'] = ['2007-01-01', '2020-09-30', 0., 2]

    df_xcats = pd.DataFrame(index=xcats,
                            columns=['earliest', 'latest', 'mean_add', 'sd_mult',
                                     'ar_coef', 'back_coef'])
    df_xcats.loc['XR'] = ['2000-01-01', '2020-12-31', 0.1, 1, 0, 0.3]
    df_xcats.loc['CRY'] = ['2000-01-01', '2020-10-30', 0, 2, 0.95, 1]
    df_xcats.loc['GROWTH'] = ['2001-01-01', '2020-10-30', 0, 2, 0.9, 1]
    df_xcats.loc['INFL'] = ['2001-01-01', '2020-10-30', 0, 2, 0.8, 0.5]

    black = {'AUD': ['2006-01-01', '2015-12-31'], 'GBP': ['2012-01-01', '2100-01-01']}

    dfd = make_qdf(df_cids, df_xcats, back_ar=0.75)

    # Additional signals.
    srn = SignalReturnRelations(dfd, ret="XR", sig="CRY", rival_sigs=None,
                                sig_neg=True, cosp=True, freq="M", start="2002-01-01")

    dfsum = srn.summary_table()

    r_sigs = ["INFL", "GROWTH"]
    srn = SignalReturnRelations(dfd, ret="XR", sig="CRY", rival_sigs=r_sigs,
                                sig_neg=True, cosp=True, freq="M", start="2002-01-01")
    dfsum = srn.summary_table()

    df_sigs = srn.signals_table(sigs=['CRY_NEG', 'INFL_NEG'])
    df_sigs_all = srn.signals_table()

    srn.accuracy_bars(type="signals", title="Accuracy measure between target return, XR,"
                                            " and the respective signals, ['CRY', 'INFL'"
                                            ", 'GROWTH'].")


"""Convergence of rows for max weight"""
import numpy as np
import warnings

class ConvergeRow(object):
    """
    Class designed to receive a row of weights, where at least one weight in the
    aforementioned row exceeds the permitted upper-bound, and subsequently redistributes
    the excess evenly across all cross-sections.
    """

    def __init__(self, row, max_weight, margin=0.001, max_loops=25):
        """
        Class's Constructor.

        :param <np.ndarray> row: Array of weights.
        :param <Float> max_weight: Maximum weight.
        :param <Float> margin: Margin of error allowed in the convergence to within the
                                upper-bound, "max_weight".
        :param <Integer> max_loops: Controls the accuracy: in theory, the greater the
                                    number of loops allowed, the more accurate the
                                    convergence. However, will only become significant if
                                    a tight margin is imposed: the "looser" the margin,
                                    the less likely the maximum number of loops permitted
                                    will be exceeded.
        """
        self.row = row
        self.max_weight = max_weight
        self.flag = (1 / np.sum(self.row > 0)) <= self.max_weight
        self.margin = margin
        self.max_loops = max_loops

    @classmethod
    def application(cls, row, max_weight):

        cr = ConvergeRow(row=row, max_weight=max_weight)
        if cr.flag:  # if enough non-nan values in row
            cr.distribute_simple()
        else:  # apply equal weights to all non-nan cases
            cr.row = (cr.row > 0) / np.sum(cr.row > 0)
            cr.row[cr.row == 0.0] = np.nan

        return cr.row

    def distribute_simple(self):
        """
        Initiates an indefinite While Loop until the weights converge below the
        (max_weight + margin). Will evenly redistribute the excess weight across all
        active cross-sections, and will compute the maximum weight, the number of cross-
        sections above the threshold and the excess weight dynamically: through each
        iteration.

        """

        count = 0
        close_enough = False
        ar_weights = self.row.copy()

        while (count <= self.max_loops) and (not close_enough):
            count += 1
            excesses = ar_weights - (self.max_weight + self.margin)
            excesses[excesses <= 0] = 0
            ar_weights = (ar_weights - excesses) + np.nanmean(excesses)

            if np.max(ar_weights) <= (self.max_weight + self.margin):
                close_enough = True

        self.row = ar_weights



import numpy as np
import pandas as pd
import seaborn as sns

import random
from typing import List, Union, Tuple
from macrosynergy.panel.historic_vol import expo_weights, expo_std, flat_std
from macrosynergy.management.shape_dfs import reduce_df_by_ticker
from macrosynergy.panel.converge_row import ConvergeRow
from macrosynergy.management.simulate_quantamental_data import make_qdf
import matplotlib.pyplot as plt
import matplotlib.dates as mdates


class Basket(object):

    def __init__(self, df: pd.DataFrame, contracts: List[str], ret: str = "XR_NSA",
                 cry: Union[str, List[str]] = None, start: str = None, end: str = None,
                 blacklist: dict = None, ewgts: List[str] = None):

        """
        Calculates the returns and carries of baskets of financial contracts using
        various weighting methods.

        :param <pd.Dataframe> df: standardized DataFrame containing the columns: 'cid',
            'xcat', 'real_date' and 'value'.
        :param <List[str]> contracts: base tickers (combinations of cross-sections and
            base categories) that define the contracts that go into the basket.
        :param <str> ret: return category postfix to be appended to the contract base;
            default is "XR_NSA".
        :param <List[str] or str> cry: carry category postfix; default is None. The field
            can either be a single carry or multiple carries defined in a List.
        :param <str> start: earliest date in ISO 8601 format. Default is None.
        :param <str> end: latest date in ISO 8601 format. Default is None.
        :param <dict> blacklist: cross-sections with date ranges that should be excluded
            from the DataFrame. If one cross-section has several blacklist periods append
            numbers to the cross-section code.
        :param List[str] ewgts: one or more postfixes that may identify exogenous weight
            categories. Similar to return postfixes they are appended to base tickers.

        N.B.: Each instance of the class will update associated standardised DataFrames,
        containing return and carry categories, and external weights.
        """
        df["real_date"] = pd.to_datetime(df["real_date"], format="%Y-%m-%d")
        df = df[["cid", "xcat", "real_date", "value"]]

        assert isinstance(contracts, list)
        c_error = "Contracts must be a list of strings."
        assert all(isinstance(c, str) for c in contracts), c_error
        assert isinstance(ret, str), "`ret` must be a string"

        df = reduce_df_by_ticker(
            df, start=start, end=end, ticks=None, blacklist=blacklist
        )
        self.contracts = contracts
        self.ret = ret
        self.ticks_ret = [con + ret for con in contracts]
        dfw_ret = self.pivot_dataframe(df, self.ticks_ret)
        self.dfw_ret = dfw_ret.dropna(axis=0, how='all')

        self.store_attributes(df, cry, "cry")
        self.store_attributes(df, ewgts, "wgt")

        self.tickers = self.ticks_ret + self.ticks_cry + self.ticks_wgt
        self.start = self.date_check(start)
        self.end = self.date_check(end)
        self.dfx = reduce_df_by_ticker(df, ticks=self.tickers)
        self.dict_retcry = {}  # dictionary for collecting basket return/carry dfs.
        self.dict_wgs = {}  # dictionary for collecting basket return/carry dfs.

    def store_attributes(self, df: pd.DataFrame, pfx: List[str], pf_name: str):
        """
        Adds multiple attributes to class based on postfixes that denote carry or
        external weight types.

        :param <pd.DataFrame> df: original, standardised DataFrame.
        :param <List[str]> pfx: category postfixes involved in the basket calculation.
        :param <str> pf_name: associated name of the postfix "cry" or "wgt".

        Note: These are [1] flags of existence of carry and weight strings in class,
        [2] lists of tickers related to all postfixes, [3] a dictionary of wide time
        series panel dataframes for all postfixes.
        """

        pfx_flag = pfx is not None
        self.__dict__[pf_name + "_flag"] = pfx_flag
        self.__dict__["ticks_" + pf_name] = []
        if pfx_flag:
            error = f"'{pf_name}' must be a <str> or a <List[str]>."
            assert isinstance(pfx, (list, str)), error
            pfx = [pfx] if isinstance(pfx, str) else pfx
            self.__dict__[pf_name] = pfx

            dfws_pfx = {}
            for cat in pfx:
                ticks = [con + cat for con in self.contracts]
                self.__dict__["ticks_" + pf_name] += ticks
                dfws_pfx[cat] = self.pivot_dataframe(df, ticks)
        else:
            dfws_pfx = None

        self.__dict__["dfws_" + pf_name] = dfws_pfx

    @staticmethod
    def pivot_dataframe(df, tick_list):
        """
        Reduces the standardised DataFrame to include a subset of the possible tickers
        and, subsequently returns a wide dataframe: each column corresponds to a ticker.

        :param <List[str]> tick_list: list of the respective tickers.
        :param <pd.DataFrame> df: standardised dataframe.

        :return <pd.DataFrame> dfw: wide dataframe.
        """

        df['ticker'] = df['cid'] + '_' + df['xcat']
        dfx = df[df["ticker"].isin(tick_list)]
        dfw = dfx.pivot(index="real_date", columns="ticker", values="value")
        return dfw

    @staticmethod
    def date_check(date_string):
        """
        Validates that the dates passed are valid timestamp expressions and will convert
        to the required form '%Y-%m-%d'. Will raise an assertion if not in the expected
        form.

        :param <str> date_string: valid date expression. For instance, "1st January,
            2000."
        """
        date_error = "Expected form of string: '%Y-%m-%d'."
        if date_string is not None:
            try:
                pd.Timestamp(date_string).strftime("%Y-%m-%d")
            except ValueError:
                raise AssertionError(date_error)
            else:
                return pd.Timestamp(date_string).strftime("%Y-%m-%d")

    @staticmethod
    def check_weights(weight: pd.DataFrame):
        """
        Checks if all rows in dataframe add up to roughly 1.

        :param <pd.DataFrame> weight: weight dataframe.
        """
        check = weight.sum(axis=1)
        c = ~((abs(check - 1) < 1e-6) | (abs(check) < 1e-6))
        assert not any(c), f"weights must sum to one (or zero), not: {check[c]}"

    @staticmethod
    def max_weight_func(weights: pd.DataFrame, max_weight: float):
        """
        Enforces maximum weight caps or - if impossible applies equal weight.

        :param <pd.DataFrame> weights: Corresponding weight matrix. Multidimensional.
        :param <float> max_weight: Upper-bound on the weight allowed for each
            cross-section.

        :return <pd.DataFrame>: Will return the modified weight DataFrame.

        N.B.: If the maximum weight is less than the equal weight weight, this replaces
        the computed weight with the equal weight. For instance,
        [np.nan, 0.63, np.nan, np.nan, 0.27] becomes [np.nan, 0.5, np.nan, np.nan, 0.5].
        Otherwise, the function calls the ConvergeRow Class to ensure all weights
        "converge" to a value within the upper-bound. Allow for a margin of error set to
        0.001.
        """

        dfw_wgs = weights.to_numpy()

        for i, row in enumerate(dfw_wgs):
            row = ConvergeRow.application(row, max_weight)
            weights.iloc[i, :] = row

        return weights

    def equal_weight(self, df_ret: pd.DataFrame) -> pd.DataFrame:
        """
        Calculates dataframe of equal weights based on available return data.

        :param <pd.DataFrame> df_ret: wide time-indexed data frame of returns.

        :return <pd.DataFrame>: dataframe of weights.

        Note: The method determines the  number of non-NA cross-sections per timestamp,
        and subsequently distributes the weights evenly across non-NA cross-sections.
        """

        act_cross = (~df_ret.isnull())
        uniform = (1 / act_cross.sum(axis=1)).values
        uniform = uniform[:, np.newaxis]

        broadcast = np.repeat(uniform, df_ret.shape[1], axis=1)

        weight = act_cross.multiply(broadcast)
        cols = weight.columns
        # Zeroes treated as NaNs.
        weight[cols] = weight[cols].replace({0: np.nan})

        self.check_weights(weight=weight)

        return weight

    def fixed_weight(self, df_ret: pd.DataFrame, weights: List[float]):
        """
        Calculates fixed weights based on a single list of values and a corresponding
        return panel dataframe.

        :param <pd.DataFrame> df_ret: Return series matrix. Multidimensional.
        :param <List[float]> weights: List of floats determining weight allocation.

        :return <pd.DataFrame>: panel of weights
        """

        act_cross = (~df_ret.isnull())

        weights = np.array(weights, dtype=np.float32)
        rows = act_cross.shape[0]
        broadcast = np.tile(weights, (rows, 1))  # Constructs Array by row repetition.

        # Replaces weight factors with zeroes if concurrent return unavailable.
        weight = act_cross.multiply(broadcast)
        cols = weight.columns
        # Zeroes treated as NaNs.
        weight[cols] = weight[cols].replace({0: np.nan})

        weight_arr = weight.to_numpy()  # convert df to np array.
        weight[weight.columns] = weight_arr / np.sum(weight_arr, axis=1)[:, np.newaxis]
        self.check_weights(weight)

        return weight

    def inverse_weight(self, dfw_ret: pd.DataFrame, lback_meth: str = "xma",
                       lback_periods: int = 21, remove_zeros: bool = True):
        """
        Calculates weights inversely proportionate to recent return standard deviations.

        :param <pd.DataFrame> dfw_ret: panel dataframe of returns.
        :param <str> lback_meth: Lookback method for "invsd" weighting method. Default is
            "xma".
        :param <int> lback_periods: Lookback periods. Default is 21.  Half-time for "xma"
            and full lookback period for "ma".
        :param <Bool> remove_zeros: Any returns that are exact zeros will not be included
            in the lookback window and prior non-zero values are added to the window
            instead.

        :return <pd.DataFrame>: Dataframe of weights.

        N.B.: The rolling standard deviation will be calculated either using the standard
        moving average (ma) or the exponential moving average (xma). Both will require
        returns before a first weight can be computed.
        """

        if lback_meth == "ma":
            dfwa = dfw_ret.rolling(window=lback_periods).agg(flat_std, remove_zeros)
            dfwa *= np.sqrt(252)

        else:

            half_life = lback_periods
            weights = expo_weights(lback_periods * 2, half_life)
            dfwa = dfw_ret.rolling(window=lback_periods * 2).agg(expo_std, w=weights,
                                                                 remove_zeros=remove_zeros)

        cols = dfwa.columns
        # Zeroes treated as NaNs.
        dfwa[cols] = dfwa[cols].replace({0: np.nan})

        df_isd = 1 / dfwa
        df_wgts = df_isd / df_isd.sum(axis=1).values[:, np.newaxis]
        self.check_weights(df_wgts)

        return df_wgts

    def values_weight(self, dfw_ret: pd.DataFrame, dfw_wgt: pd.DataFrame,
                      weight_meth: str):
        """
        Returns weights based on an external weighting category.

        :param <pd.DataFrame> dfw_ret: Standard wide dataframe of returns across time and
            contracts.
        :param <pd.DataFrame> dfw_wgt: Standard wide dataframe of weight category values
            across time and contracts.
        :param <str> weight_meth:

        :return <pd.DataFrame>: Dataframe of weights.
        """

        negative_condition = np.any((dfw_wgt < 0).to_numpy())
        if negative_condition:
            dfw_wgt[dfw_wgt < 0] = 0.0
            print("Negative values in the weight matrix set to zero.")

        exo_array = dfw_wgt.to_numpy()
        df_bool = ~dfw_ret.isnull()

        weights_df = df_bool.multiply(exo_array)
        cols = weights_df.columns

        # Zeroes treated as NaNs.
        weights_df[cols] = weights_df[cols].replace({0: np.nan})

        if weight_meth != "values":
            weights_df = 1 / weights_df

        weights = weights_df.divide(weights_df.sum(axis=1), axis=0)
        self.check_weights(weights)

        return weights

    def make_weights(self, weight_meth: str = "equal", weights: List[float] = None,
                     lback_meth: str = "xma", lback_periods: int = 21,
                     ewgt: str = None, max_weight: float = 1.0,
                     remove_zeros: bool = True):
        """
        Returns wide dataframe of weights to be used for basket series.

        :param <str> weight_meth: method used for weighting constituent returns and
            carry. The parameter can receive either a single weight method or
            multiple weighting methods. See `make_basket` docstring.
        :param <List[float]> weights: single list of weights corresponding to the base
            tickers in `contracts` argument. This is only relevant for the fixed weight
            method.
        :param <str> lback_meth: look-back method for "invsd" weighting method. Default
            is Exponential MA, "ema". The alternative is simple moving average, "ma".
        :param <int> lback_periods: look-back periods for "invsd" weighting method.
            Default is 21.  Half-time for "xma" and full lookback period for "ma".
        :param <str> ewgt: Exogenous weight postfix that defines the weight value panel.
            Only needed for the 'values' or 'inv_values' method.
        :param <float> max_weight: maximum weight of a single contract. Default is 1, i.e
            zero restrictions. The purpose of the restriction is to limit concentration
            within the basket.
        :param <bool> remove_zeros: removes the zeros. Default is set to True.

        return: <pd.DataFrame>: wide dataframe of contract weights across time.
        """
        assert 0.0 < max_weight <= 1.0
        assert weight_meth in ['equal', 'fixed', 'values', 'inv_values', 'invsd']

        # Apply weight method.

        if weight_meth == "equal":
            dfw_wgs = self.equal_weight(df_ret=self.dfw_ret)

        elif weight_meth == "fixed":
            message = "Expects a list of weights."
            message_2 = "List of weights must be equal to the number of contracts."
            assert isinstance(weights, list), message
            assert self.dfw_ret.shape[1] == len(weights), message_2
            message_3 = "Expects a list of floating point values."
            assert all(isinstance(w, (int, float)) for w in weights), message_3

            dfw_wgs = self.fixed_weight(df_ret=self.dfw_ret, weights=weights)

        elif weight_meth == "invsd":
            error_message = "Lookback method method must be 'ma' or 'xma'."
            assert lback_meth in ["xma", "ma"], error_message
            assert isinstance(lback_periods, int), "Expects <int>."
            dfw_wgs = self.inverse_weight(dfw_ret=self.dfw_ret, lback_meth=lback_meth,
                                          lback_periods=lback_periods,
                                          remove_zeros=remove_zeros)

        elif weight_meth in ["values", "inv_values"]:
            assert ewgt in self.wgt, f'{ewgt} is not defined on the instance.'
            # Lag by one day to be used as weights.
            try:
                dfw_wgt = self.dfws_wgt[ewgt].shift(1)
            except KeyError as e:
                print(f"Basket not found: {e}.")
            else:
                cols = sorted(dfw_wgt.columns)
                dfw_ret = dfw_wgt.reindex(cols, axis=1)
                dfw_wgt = dfw_wgt.reindex(cols, axis=1)
                dfw_wgs = self.values_weight(dfw_ret, dfw_wgt, weight_meth)

        else:
            raise NotImplementedError(f"Weight method unknown {weight_meth}")

        # Remove leading NA rows.

        fvi = max(dfw_wgs.first_valid_index(), self.dfw_ret.first_valid_index())
        dfw_wgs = dfw_wgs[fvi:]

        # Impose cap on cross-section weights.

        if max_weight < 1.0:
            dfw_wgs = self.max_weight_func(weights=dfw_wgs, max_weight=max_weight)

        return dfw_wgs

    @staticmethod
    def column_manager(df_cat: pd.DataFrame, dfw_wgs: pd.DataFrame):
        """
        Will match the column names of the two dataframes involved in the computation:
        either the return & weight dataframes or the carry & weight dataframes. The
        pandas multiply operation requires the column names, of both dataframes involved
        in the binary operation, to be identical.

        :param <pd.DataFrame> df_cat: return or carry dataframe.
        :param <pd.DataFrame> dfw_wgs: weight dataframe.

        :return <pd.DataFrame> dfw_wgs: modified weight dataframe (column names will map
            to the other dataframe received).
        """

        df_cat = df_cat.reindex(sorted(df_cat.columns), axis=1)
        dfw_wgs = dfw_wgs.reindex(sorted(dfw_wgs.columns), axis=1)

        ret_cols = df_cat.columns
        weight_cols = dfw_wgs.columns

        if all(ret_cols != weight_cols):
            dfw_wgs.columns = ret_cols

        return dfw_wgs

    def column_weights(self, dfw_wgs: pd.DataFrame):
        """
        The weight dataframe is used to compute the basket performance for returns,
        carries etc. Therefore, with their broad application, the column names of the
        dataframe should correspond to the ticker postfix of each contract.

        :param <pd.DataFrame> dfw_wgs: weight dataframe.

        :return <pd.DataFrame> dfw_wgs: weight dataframe with updated columns names.
        """

        dfw_weight_names = lambda w_name: w_name[:w_name.find(self.w_field)]
        if self.wgt_flag and self.exo_w_postfix is not None:
            self.__dict__['w_field'] = self.exo_w_postfix
        else:
            self.__dict__['w_field'] = self.ret

        cols = list(map(dfw_weight_names, dfw_wgs.columns))
        dfw_wgs.columns = cols
        dfw_wgs.columns.name = "ticker"

        return dfw_wgs

    def make_basket(self, weight_meth: str = "equal", weights: List[float] = None,
                    lback_meth: str = "xma", lback_periods: int = 21,
                    ewgt: str = None, max_weight: float = 1.0, remove_zeros: bool = True,
                    basket_name: str = "GLB_ALL"):
        """
        Calculates all basket performance categories.

        :param <str> weight_meth: method used for weighting constituent returns and
            carry. The parameter can receive either a single weight method or
            multiple weighting methods. The options are as follows:
            [1] "equal": all constituents with non-NA returns have the same weight.
                This is the default.
            [2] "fixed": weights are proportionate to a single list of values provided
                which are passed to argument `weights` (each value corresponds to a
                single contract).
            [3] "invsd": weights based on inverse to standard deviations of recent
                returns.
            [4] "values": weights proportionate to a panel of values of exogenous weight
                category.
            [5] "inv_values": weights are inversely proportionate to of values of exogenous
                weight category.
        :param <List[float]> weights: single list of weights corresponding to the base
            tickers in `contracts` argument. This is only relevant for the fixed weight
            method.
        :param <str> lback_meth: look-back method for "invsd" weighting method. Default
            is Exponential MA, "ema". The alternative is simple moving average, "ma".
        :param <int> lback_periods: look-back periods for "invsd" weighting method.
            Default is 21.  Half-time for "xma" and full lookback period for "ma".
        :param <str> ewgt: Exogenous weight postfix that defines the weight value panel.
            Only needed for the 'values' or 'inv_values' method.
        :param <float> max_weight: maximum weight of a single contract. Default is 1, i.e
            zero restrictions. The purpose of the restriction is to limit concentration
            within the basket.
        :param <bool> remove_zeros: removes the zeros. Default is set to True.
        :param <str> basket_name: name of basket base ticker (analogous to contract name)
            to be used for return and (possibly) carry are calculated. Default is
            "GLB_ALL".
        """

        assert isinstance(weight_meth, str), "`weight_meth` must be string"

        self.__dict__['exo_w_postfix'] = ewgt
        dfw_wgs = self.make_weights(weight_meth=weight_meth, weights=weights,
                                    lback_meth=lback_meth, lback_periods=lback_periods,
                                    ewgt=ewgt, max_weight=max_weight,
                                    remove_zeros=remove_zeros)
        select = ["ticker", "real_date", "value"]

        dfw_wgs_copy = self.column_manager(df_cat=self.dfw_ret, dfw_wgs=dfw_wgs)

        dfw_bret = self.dfw_ret.multiply(dfw_wgs_copy).sum(axis=1)
        dfxr = dfw_bret.to_frame("value").reset_index()
        basket_ret = basket_name + "_" + self.ret
        dfxr = dfxr.assign(ticker=basket_ret)[select]
        store = [dfxr]

        if self.cry_flag:
            cry_list = []
            for cr in self.cry:
                dfw_wgs_copy = self.column_manager(df_cat=self.dfws_cry[cr],
                                                   dfw_wgs=dfw_wgs)
                dfw_bcry = self.dfws_cry[cr].multiply(dfw_wgs_copy).sum(axis=1)
                dfcry = dfw_bcry.to_frame("value").reset_index()
                basket_cry = basket_name + "_" + cr
                dfcry = dfcry.assign(ticker=basket_cry)[select]
                cry_list.append(dfcry)

            store += cry_list

        df_retcry = pd.concat(store)
        df_retcry = df_retcry.reset_index(drop=True)
        self.dict_retcry[basket_name] = df_retcry

        self.dict_wgs[basket_name] = self.column_weights(dfw_wgs)

    def weight_visualiser(self, basket_name, start_date: str = None,
                          end_date: str = None, subplots: bool = True,
                          facet_grid: bool = False, scatter: bool = False,
                          all_tickers: bool = True, single_ticker: str = None,
                          percentage_change: bool = False, 
                          size: Tuple[int, int] = (7, 7),):
        """
        Method used to visualise the weights associated with each contract in the basket.

        :param <str> basket_name: name of basket whose weights are visualized
        :param <str> start_date: start date of he visualisation period.
        :param <str> end_date: end date of the visualization period.
        :param <bool> subplots: contract weights are displayed on different plots (True)
            or on a single plot (False).
        :param <bool> facet_grid: parameter used to break up the plot into multiple
            cartesian coordinate systems. If the basket consists of a high number of
            contracts, using the Facet Grid is recommended.
        :param <bool> scatter: if the facet_grid parameter is set to True there are two
            options: i) scatter plot if there a lot of blacklist periods; ii) line plot
            for continuous series.
        :param <bool> all_tickers: if True (default) all weights are displayed.
            If set to False `single-ticker` must be specified.
        :param <str> single_ticker: individual ticker for further, more detailed,
            analysis.
        :param <bool> percentage_change: graphical display used to further assimilate the
            fluctuations in the contract's weight. The graphical display is limited to a
            single contract. Therefore, pass the ticker into the parameter
            "single_ticker".
        :param <Tuple[int, int]> size: size of the plot. Default is (7, 7).

        """

        date_conv = lambda d: pd.Timestamp(d).strftime("%Y-%m-%d %X")
        try:
            dfw_wgs = self.dict_wgs[basket_name]
        except KeyError as e:
            print(f"Basket not found - call make_basket() method first: {e}.")
        else:
            if isinstance(start_date, str) and isinstance(end_date, str):
                self.date_check(start_date)
                self.date_check(end_date)
                start_date = date_conv(start_date)
                end_date = date_conv(end_date)
            elif isinstance(start_date, str):
                self.date_check(start_date)
                start_date = date_conv(start_date)
                end_date = dfw_wgs.index[-1]
            elif isinstance(end_date, str):
                self.date_check(end_date)
                start_date = dfw_wgs.index[0]
                end_date = date_conv(end_date)
            else:
                start_date = dfw_wgs.index[0]
                end_date = dfw_wgs.index[-1]

            error_1 = f"{start_date} unavailable in weight dataframe."
            c = dfw_wgs.index
            assert start_date in c, error_1
            error_2 = f"{end_date} unavailable in weight dataframe."
            assert end_date in c, error_2

            dfw_wgs = dfw_wgs.truncate(before=start_date, after=end_date)
            if not all_tickers:
                error_3 = "The parameter, 'single_ticker', must be a <str>."
                assert isinstance(single_ticker, str), error_3
                error_4 = f"Ticker not present in the weight dataframe. Available " \
                          f"tickers: {list(dfw_wgs.columns)}."
                assert single_ticker in dfw_wgs.columns, error_4
                dfw_wgs = dfw_wgs[[single_ticker]]

            if facet_grid:
                df_stack = dfw_wgs.stack().to_frame("value").reset_index()
                df_stack = df_stack.sort_values(['ticker', 'real_date'])
                no_contracts = dfw_wgs.shape[1]
                facet_cols = 4 if no_contracts >= 8 else 3
                sns.set(rc={'figure.figsize': size})
                fg = sns.FacetGrid(df_stack, col="ticker", col_wrap=facet_cols,
                                   sharey=True)

                scatter_error = f"Boolean object expected - instead received " \
                                f"{type(scatter)}."
                assert isinstance(scatter, bool), scatter_error
                if scatter:
                    fg.map_dataframe(sns.scatterplot, x='real_date',
                                     y='value')
                else:
                    # Seaborn will linearly interpolate NaN values which is visually
                    # misleading. Therefore, aim to negate the operation.
                    fg.map_dataframe(sns.lineplot, x='real_date', y='value',
                                    hue=df_stack['value'].isna().cumsum(),
                                    palette=["blue"] *
                                            df_stack['value'].isna().cumsum().nunique(),
                                    estimator=None, markers=True)

                equal_value = (1 / no_contracts)
                fg.map(plt.axhline, y=equal_value, linestyle='--', color='gray', lw=0.5)
                # Set axes labels of individual charts.
                fg.set_axis_labels('', '')
                fg.set_titles(col_template='{col_name}')
                fg.fig.suptitle('Contract weights in basket', y=1.02)
            else:
                plt.rcParams["figure.figsize"] = size
                dfw_wgs.plot(subplots=subplots, title="Weight Values Timestamp",
                             legend=True)
                plt.xlabel('real_date, years')

            date_func = lambda d: pd.Timestamp(d).strftime("%Y-%m-%d")
            if percentage_change:
                error_5 = "Percentage change display is applied to a single ticker. Set " \
                          "the parameter 'all_tickers' to False."
                assert dfw_wgs.shape[1] == 1, error_5

                plt.rcParams["figure.figsize"] = size
                fig, ax = plt.subplots()
                dfw_pct = dfw_wgs.pct_change(periods=1) * 100
                n_index = np.array(list(map(date_func, dfw_pct.index)))
                dfw_pct = dfw_pct.set_index(keys=n_index)
                dfw_pct.plot(kind='bar', color='coral', ax=ax)
                ax.xaxis.set_major_locator(mdates.MonthLocator())
                plt.xticks(rotation=0)
                ax.set_ylabel("Percentage Change in weight.")
                ax.legend()

            plt.show()

    @staticmethod
    def column_split(df: pd.DataFrame):
        """
        Receives a dataframe with the columns 'ticker', 'real_date' and 'value' and
        returns a standardised dataframe with the columns 'cid', 'xcat', 'real_date' and
        'value. The 'ticker' column is broken up to produce the two new columns.

        :param <pd.DataFrame> df:

        :return <pd.DataFrame> df: standardised dataframe.
        """
        select = ["cid", "xcat", "real_date", "value"]

        cid_func = lambda t: t.split('_')[0]
        xcat_func = lambda t: '_'.join(t.split('_')[1:])

        cids_w_df = list(map(cid_func, df["ticker"]))
        df["cid"] = np.array(cids_w_df)

        df = df.rename(columns={"ticker": "xcat"})
        df["xcat"] = np.array(list(map(xcat_func, df["xcat"])))

        df = df[select]
        return df

    def return_basket(self, basket_names: Union[str, List[str]] = None):
        """
        Return standardized dataframe of basket performance data based on one or more
        weighting methods.

        :param <str or List[str]> basket_names: single basket name or list for which
            performance data are to be returned. If none is given all baskets added to
            the instance are selected.

        :return <pd.Dataframe>: standardized DataFrame with the basket return and
            (possibly) carry data in standard form, i.e. columns 'cid', 'xcats',
            'real_date' and 'value'.
        """

        if basket_names is None:
            basket_names = list(self.dict_retcry.keys())

        basket_error = "String or List of basket names expected."
        assert isinstance(basket_error, (list, str)), basket_error
        if isinstance(basket_names, str):
            basket_names = [basket_names]

        ret_baskets = []
        for b in basket_names:
            try:
                dfw_retcry = self.dict_retcry[b]
            except KeyError as e:
                print(f"Basket not found - call make_basket() method first: {e}.")
            else:
                dfw_retcry = self.column_split(dfw_retcry)
                dfw_retcry = dfw_retcry.sort_values(['xcat', 'real_date'])
                ret_baskets.append(dfw_retcry)

        return_df = pd.concat(ret_baskets)
        return return_df.reset_index(drop=True)

    def return_weights(self, basket_names: Union[str, List[str]] = None):
        """
        Return the standardised dataframe containing the corresponding weights used to
        compute the basket.

        :param <str or List[str]> basket_names: single basket name or list for which
            performance data are to be returned. If none is given all baskets added to
            the instance are selected.

        :return <pd.Dataframe>: standardized DataFrame with basket weights.
        """

        if basket_names is None:
            basket_names = list(self.dict_wgs.keys())

        basket_error = "String or List of basket names expected."
        assert isinstance(basket_error, (list, str)), basket_error
        if isinstance(basket_names, str):
            basket_names = [basket_names]

        weight_baskets = []
        select = ["cid", "xcat", "real_date", "value"]

        for b in basket_names:
            try:
                dfw_wgs = self.dict_wgs[b]
            except KeyError as e:
                print(f"Basket not found - call make_basket() method first: {e}.")
            else:

                w = dfw_wgs.stack().to_frame("value").reset_index()
                w = self.column_split(df=w)
                w = w.sort_values(['cid', 'real_date'])

                w = w.loc[w.value > 0, select]
                w["xcat"] += "_" + b + "_" + "WGT"
                weight_baskets.append(w)

        return_df = pd.concat(weight_baskets)
        return return_df.reset_index(drop=True)


if __name__ == "__main__":

    cids = ['AUD', 'GBP', 'NZD', 'USD']
    xcats = ['FXXR_NSA', 'FXCRY_NSA', 'FXCRR_NSA', 'EQXR_NSA', 'EQCRY_NSA', 'EQCRR_NSA',
             'FXWBASE_NSA', 'EQWBASE_NSA']

    df_cids = pd.DataFrame(index=cids, columns=['earliest', 'latest', 'mean_add',
                                                'sd_mult'])

    df_cids.loc['AUD'] = ['2000-01-01', '2022-03-14', 0, 1]
    df_cids.loc['GBP'] = ['2001-01-01', '2022-03-14', 0, 2]
    df_cids.loc['NZD'] = ['2002-01-01', '2022-03-14', 0, 3]
    df_cids.loc['USD'] = ['2000-01-01', '2022-03-14', 0, 4]

    df_xcats = pd.DataFrame(index=xcats, columns=['earliest', 'latest', 'mean_add',
                                                  'sd_mult', 'ar_coef', 'back_coef'])
    df_xcats.loc['FXXR_NSA'] = ['2010-01-01', '2022-03-14', 0, 1, 0, 0.2]
    df_xcats.loc['FXCRY_NSA'] = ['2010-01-01', '2022-03-14', 1, 1, 0.9, 0.2]
    df_xcats.loc['FXCRR_NSA'] = ['2010-01-01', '2022-03-14', 0.5, 0.8, 0.9, 0.2]
    df_xcats.loc['EQXR_NSA'] = ['2010-01-01', '2022-03-14', 0.5, 2, 0, 0.2]
    df_xcats.loc['EQCRY_NSA'] = ['2010-01-01', '2022-03-14', 2, 1.5, 0.9, 0.5]
    df_xcats.loc['EQCRR_NSA'] = ['2010-01-01', '2022-03-14', 1.5, 1.5, 0.9, 0.5]
    df_xcats.loc['FXWBASE_NSA'] = ['2010-01-01', '2022-02-01', 1, 1.5, 0.8, 0.5]
    df_xcats.loc['EQWBASE_NSA'] = ['2010-01-01', '2022-02-01', 1, 1.5, 0.9, 0.5]

    random.seed(2)
    dfd = make_qdf(df_cids, df_xcats, back_ar=0.75)

    black = {'AUD': ['2010-01-01', '2013-12-31'], 'GBP': ['2010-01-01', '2013-12-31']}
    contracts = ['AUD_FX', 'AUD_EQ', 'NZD_FX', 'GBP_EQ', 'USD_EQ']
    gdp_figures = [17.0, 17.0, 41.0, 9.0, 250.0]

    contracts_1 = ['AUD_FX', 'GBP_FX', 'NZD_FX', 'USD_EQ']

    # First test. Multiple carries. Equal weight method.
    # The main aspect to check in the code is that basket performance has been applied to
    # both the return category and the multiple carry categories.
    dfd["grading"] = np.ones(dfd.shape[0])

    basket_1 = Basket(
        df=dfd, contracts=contracts_1, ret="XR_NSA", cry=["CRY_NSA", "CRR_NSA"],
        blacklist=black
    )
    basket_1.make_basket(
        weight_meth="equal", max_weight=0.55, basket_name="GLB_EQUAL"
    )

    basket_1.make_basket(
        weight_meth="fixed", max_weight=0.55, weights=[1/6, 1/6, 1/6, 1/2],
        basket_name="GLB_FIXED"
    )

    # show the weights of the GLB_FIXED basket

    basket_1.weight_visualiser(
        basket_name="GLB_FIXED",
        subplots=False, size=(10, 5)
    )



import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
import seaborn as sns
from typing import List, Union, Tuple
from scipy import stats
import statsmodels.api as sm
import warnings

from macrosynergy.management.simulate_quantamental_data import make_qdf
from macrosynergy.management.shape_dfs import categories_df


class CategoryRelations(object):
    """
    Class for analyzing and visualizing the relations of two panel categories.

    :param <pd.DataFrame> df: standardized DataFrame with the necessary columns:
        'cid', 'xcat', 'real_date' and at least one column with values of interest.
    :param <List[str]> xcats: exactly two extended categories to be analyzed.
        If there is a hypothesized explanatory-dependent relation, the first category
        is the explanatory variable and the second category the explained variable.
    :param <List[str]> cids: cross-sections for which the category relations is being
        analyzed. Default is all in the DataFrame.
    :param <str> start: earliest date in ISO format. Default is None in which case the
        earliest date in the DataFrame will be used.
    :param <str> end: latest date in ISO format. Default is None in which case the
        latest date in the DataFrame will be used.
    :param <dict> blacklist: cross-sections with date ranges that should be excluded from
        the analysis.
    :param <int> years: number of years over which data are aggregated. Supersedes the
        'freq' parameter and does not allow lags, Default is None, meaning no multi-year
        aggregation.
        Note: for single year labelled plots, better use freq='A' for cleaner labels.
    :param <str> val: name of column that contains the values of interest. Default is
        'value'.
    :param <str> freq: letter denoting frequency at which the series are to be sampled.
        This must be one of 'D', 'W', 'M', 'Q', 'A'. Default is 'M'.
    :param <int> lag: lag (delay of arrival) of first (explanatory) category in periods
        as set by freq. Default is 0.
        Importantly, for analyses with explanatory and dependent categories, the first
        category takes the role of the explanatory and a positive lag means that the
        explanatory values will be deferred into the future, i.e. relate to future values
        of the explained variable.
    :param <List[str]> xcat_aggs: Exactly two aggregation methods. Default is 'mean' for
        both.
    :param <str> xcat1_chg: time series changes are applied to the first category.
        Default is None. Options are 'diff' (first difference) and 'pch'
        (percentage change). The changes are calculated over the number of
        periods determined by `n_periods`.
    :param <int> n_periods: number of periods over which changes of the first category
        have been calculated. Default is 1.
    :param <int> fwin: forward moving average window of second category. Default is 1,
        i.e no average.
        Importantly, for analysis with explanatory and dependent categories, the second
        takes the role of the dependent and a forward window means that the dependent
        values average forward into the future.
    :param: <List[float]> xcat_trims: two-element list with maximum absolute values
        for the two respective categories. Observations with higher values will be
        trimmed, i.e. removed from the analysis (not winsorized!). Default is None
        for both. Trimming is applied after all other transformations.
    :param <int> slip: implied slippage of feature availability for relationship with
        the target category. This mimics the relationship between trading signals and
        returns, which is often characterized by a delay due to the setup of of positions.
        Technically, this is a negative lag (early arrival) of the target category
        in working days prior to any frequency conversion. Default is 0.
    """
    
    def __init__(self, df: pd.DataFrame, xcats: List[str], cids: List[str] = None,
                 val: str = 'value', start: str = None, end: str = None,
                 blacklist: dict = None, years: int = None, freq: str = 'M', lag: int = 0,
                 fwin: int = 1, xcat_aggs: List[str] = ['mean', 'mean'],
                 xcat1_chg: str = None, n_periods: int = 1,
                 xcat_trims: List[float] = [None, None], slip: int = 0,):
        """ Initializes CategoryRelations """

        self.xcats: List[str] = xcats
        self.cids: List[str] = cids 
        self.val: str = val 
        self.freq: str = freq.upper()
        self.lag: int = lag
        self.years: int  = years 
        self.aggs: List[str] = xcat_aggs
        self.xcat1_chg: str  = xcat1_chg
        self.n_periods: int = n_periods
        self.xcat_trims: List[float] = xcat_trims
        self.slip: int = slip

        if not isinstance(self.freq, str):
            raise TypeError("freq must be a string.")
        elif  self.freq not in ['D', 'W', 'M', 'Q', 'A']:
            raise ValueError("freq must be one of 'D', 'W', 'M', 'Q', 'A'.")
        if not isinstance(val, str):
            raise TypeError("val must be a string.")
        if not {'cid', 'xcat', 'real_date', val}.issubset(set(df.columns)):
            raise ValueError("`df` must have columns 'cid', 'xcat', 'real_date' and `val`.")
        if not isinstance(xcats, (list, tuple)):
            raise TypeError("`xcats` must be a list or a tuple.")
        elif not len(xcats) == 2:
            raise ValueError("`xcats` must have exactly two elements.")
        if not isinstance(slip, int):
            raise TypeError("`slip` must be a non-negative integer.")
        elif slip < 0:
            raise ValueError("`slip` must be a non-negative integer.")

        if not isinstance(xcat_aggs, (list, tuple)):
            raise TypeError("xcat_aggs must be a list or a tuple.")

        # Select the cross-sections available for both categories.
        df["real_date"] = pd.to_datetime(df["real_date"], format="%Y-%m-%d")
        
        if self.slip != 0:
            metrics_found : List[str] = list(set(df.columns) - set(['cid', 'xcat', 'real_date']))
            df = self.apply_slip(target_df=df, slip=self.slip, cids=self.cids,
                                        xcats=self.xcats, metrics=metrics_found)

        shared_cids = CategoryRelations.intersection_cids(df, xcats, cids)
        
        # Will potentially contain NaN values if the two categories are defined over
        # time-periods.
        df = categories_df(
            df, xcats, shared_cids, val=val, start=start, end=end, freq=freq,
            blacklist=blacklist, years=years, lag=lag,
            fwin=fwin, xcat_aggs=xcat_aggs
        )

        if xcat1_chg is not None:

            xcat1_error = "Change applied to the explanatory variable must either be " \
                          "first-differencing, 'diff', or percentage change, 'pch'."
            assert xcat1_chg in ['diff', 'pch'], xcat1_error
            n_periods_error = f"<int> expected and not {type(n_periods)}."
            assert isinstance(n_periods, int), n_periods_error

            df = CategoryRelations.time_series(
                df, change=xcat1_chg, n_periods=n_periods, shared_cids=shared_cids,
                expln_var=xcats[0]
            )

        if any([xt is not None for xt in self.xcat_trims]):

            xcat_trim_error = "Two values expected corresponding to the number " \
                              "of categories."
            assert len(xcat_trims) == len(xcats), xcat_trim_error

            types = [isinstance(elem, (float, int)) and elem >= 0.0
                     for elem in xcat_trims]
            assert any(types), "Expected two floating point values."

            df = CategoryRelations.outlier_trim(df, xcats, xcat_trims)

        # NaN values will not be handled if both of the above conditions are not
        # satisfied.
        self.df = df.dropna(axis=0, how='any')

    @classmethod
    def apply_slip(self, target_df: pd.DataFrame, slip: int,
                    cids: List[str], xcats: List[str],
                    metrics: List[str]) -> pd.DataFrame:
        """
        Applied a slip, i.e. a negative lag, to the target DataFrame 
        for the given cross-sections and categories, on the given metrics.
        
        Parameters
        ----------
        :param <pd.DataFrame> target_df: DataFrame to which the slip is applied.
        :param <int> slip: Slip to be applied.
        :param <List[str]> cids: List of cross-sections.
        :param <List[str]> xcats: List of categories.
        :param <List[str]> metrics: List of metrics to which the slip is applied.
        :return <pd.DataFrame> target_df: DataFrame with the slip applied.
        :raises <TypeError>: If the provided parameters are not of the expected type.
        :raises <ValueError>: If the provided parameters are semantically incorrect.
        """

        target_df = target_df.copy(deep=True)
        if not (isinstance(slip, int) and slip >= 0):
            raise ValueError("Slip must be a non-negative integer.")

        sel_tickers : List[str] = [f"{cid}_{xcat}" for cid in cids for xcat in xcats]
        target_df['tickers'] = target_df['cid'] + '_' + target_df['xcat']

        if not set(sel_tickers).issubset(set(target_df['tickers'].unique())):
            raise ValueError("Tickers targetted for applying slip are not present in the DataFrame.\n"
             f"Missing tickers: {set(sel_tickers) - set(target_df['tickers'].unique())}")

        slip : int = slip.__neg__()
        
        target_df[metrics] = target_df.groupby('tickers')[metrics].shift(slip)
        target_df = target_df.drop(columns=['tickers'])
        
        return target_df
        


    @classmethod
    def intersection_cids(cls, df, xcats, cids):
        """Returns common cross-sections across both categories and specified
        parameter.

        :param <pd.DataFrame> df: standardised DataFrame.
        :param <List[str]> xcats: exactly two extended categories to be checked on.
        :param <List[str]> cids: cross-sections for which the category relation is being
        analyzed.

        :return <List[str]>: usable: List of the common cross-sections across the two
            categories.
        """

        set_1 = set(df[df['xcat'] == xcats[0]]['cid'])
        set_2 = set(df[df['xcat'] == xcats[1]]['cid'])

        miss_1 = list(set(cids).difference(set_1))
        miss_2 = list(set(cids).difference(set_2))

        if len(miss_1) > 0:
            print(f"{xcats[0]} misses: {sorted(miss_1)}.")
            warnings.warn(f"{xcats[0]} misses: {sorted(miss_1)}.", UserWarning)
        if len(miss_2) > 0:
            print(f"{xcats[1]} misses: {sorted(miss_2)}.")
            warnings.warn(f"{xcats[1]} misses: {sorted(miss_2)}.", UserWarning)

        usable = list(set_1.intersection(set_2).
                      intersection(set(cids)))

        return usable

    @classmethod
    def time_series(cls, df: pd.DataFrame, change: str, n_periods: int,
                    shared_cids: List[str], expln_var: str):
        """ Calculates first differences and percent changes.

        :param <pd.DataFrame> df: multi-index DataFrame hosting the two categories: first
            column represents the explanatory variable; second column hosts the dependent
            variable. The DataFrame's index is the real-date and cross-section.
        :param <str> change: type of change to be applied
        :param <int> n_periods: number of base periods in df over which the change is applied.
        :param <List[str]> shared_cids: shared cross-sections across the two categories
            and the received list.
        :param <str> expln_var: only the explanatory variable's data series will be
            changed from the raw value series to a difference or percentage change value.

        :return <pd.Dataframe>: df: returns the same multi-index DataFrame but with an
            adjusted series inline with the 'change' parameter.
        """

        df_lists = []
        for c in shared_cids:
            temp_df = df.loc[c]

            if change == 'diff':
                temp_df[expln_var] = temp_df[expln_var].diff(periods=n_periods)
            else:
                temp_df[expln_var] = temp_df[expln_var].pct_change(periods=n_periods)

            temp_df['cid'] = c
            temp_df = temp_df.set_index('cid', append=True)
            df_lists.append(temp_df)

        df_ = pd.concat(df_lists)
        df_ = df_.dropna(axis=0, how='any')
        return df_

    @classmethod
    def outlier_trim(cls, df: pd.DataFrame, xcats: List[str], xcat_trims: List[float]):
        """
        Trim outliers from the dataset.

        :param <pd.DataFrame> df: multi-index DataFrame hosting the two categories. The
            transformations, to each series, have already been applied.
        :param <List[str]> xcats: explanatory and dependent variable.
        :param <List[float]> xcat_trims:

        :return <pd.DataFrame> df: returns the same multi-index DataFrame.

        N.B.:
        Outliers are classified as any datapoint whose absolute value exceeds the
        predefined value specified in the field self.xcat_trims. The values will be set
        to NaN, and subsequently excluded from any regression modelling or correlation
        coefficients.
        """

        xcat_dict = dict(zip(xcats, xcat_trims))

        for k, v in xcat_dict.items():

            df[k] = np.where(np.abs(df[k]) < v, df[k], np.nan)

        df = df.dropna(axis=0, how='any')
        return df

    def corr_prob_calc(self, df_probability: Union[pd.DataFrame, List[pd.DataFrame]],
                       prob_est):
        """
        Compute the correlation coefficient and probability statistics.

        :param <List[pd.DataFrame] or pd.DataFrame> df_probability: pandas DataFrame
            containing the dependent and explanatory variables.
        :param <bool> prob_bool: boolean parameter which determines whether the
            probability value is included in the table. The default is True.

        :return <List[tuple(float, float)]>:

        N.B.: The method is able to handle multiple DataFrames, and will return the
        corresponding number of statistics held inside a List.
        """
        if isinstance(df_probability, pd.DataFrame):
            df_probability = [df_probability]

        cpl = []
        for i, df_i in enumerate(df_probability):
            feat = df_i[self.xcats[0]].to_numpy()
            targ = df_i[self.xcats[1]].to_numpy()
            coeff, pval = stats.pearsonr(feat, targ)
            if prob_est == "map":
                X = df_i.loc[:, self.xcats[0]]
                X = sm.add_constant(X)
                y = df_i.loc[:, self.xcats[1]]
                groups = df_i.reset_index().real_date
                re = sm.MixedLM(y, X, groups, ).fit(reml=False)  # random effects est
                pval = float(re.summary().tables[1].iloc[1, 3])
            row = [np.round(coeff, 3), np.round(1 - pval, 3)]
            cpl.append(row)
        return cpl

    def corr_probability(self, df_probability, prob_est, time_period: str = '',
                         coef_box_loc: str = 'upper left'):
        """
        Add the computed correlation coefficient and probability to a Matplotlib table.

        :param <List[pd.DataFrame] or pd.DataFrame> df_probability: pandas DataFrame
            containing the dependent and explanatory variables. Able to handle multiple
            DataFrames representing different time-periods of the original series.
        :param <str> time_period: indicator used to clarify which time-period the
            statistics are computed for. For example, before 2010 and after 2010: the two
            periods experience very different macroeconomic conditions. The default is
            an empty string.
        :param <str> coef_box_loc: location on the graph of the aforementioned box. The
            default is in the upper left corner.
        :param <bool> prob_bool: boolean parameter which determines whether the
            probability value is included in the table. The default is True.

        """
        time_period_error = f"<str> expected - received {type(time_period)}."
        assert isinstance(time_period, str), time_period_error

        cpl = self.corr_prob_calc(df_probability=df_probability,
                                  prob_est=prob_est)

        fields = [f"Correlation\n coefficient {time_period}",
                  f"Probability\n of significance {time_period}"]

        if isinstance(df_probability, list) and len(df_probability) == 2:
            row_headers = ["Before 2010", "After 2010"]
            cellC = [["lightsteelblue", "lightsteelblue"],
                     ["lightsalmon", "lightsalmon"]]
        else:
            row_headers = None
            cellC = None

        data_table = plt.table(cellText=cpl, cellColours=cellC,
                               colLabels=fields, cellLoc='center', loc=coef_box_loc)
        
        return data_table

    def annotate_facet(self, data, **kws):
        """ Annotate each graph within the facet grid. """

        x = data[self.xcats[0]].to_numpy()
        y = data[self.xcats[1]].to_numpy()
        coeff, pval = stats.pearsonr(x, y)

        cpl = np.round(coeff, 3)
        fields = "Correlation coefficient: "
        ax = plt.gca()
        ax.text(.04, .1, f"{fields} {cpl}", fontsize=10, transform=ax.transAxes)

    def reg_scatter(self, title: str = None, labels: bool = False,
                    size: Tuple[float] = (12, 8), xlab: str = None, ylab: str = None,
                    coef_box: str = None, prob_est: str = "pool",
                    fit_reg: bool = True, reg_ci: int = 95,
                    reg_order: int = 1, reg_robust: bool = False,
                    separator: Union[str, int] = None, title_adj: float = 1,
                    single_chart: bool = False):

        """
        Display scatter-plot and regression line.

        :param <str> title: title of plot. If None (default) an informative title is
            applied.
        :param <bool> labels: assign a cross-section/period label to each dot.
            Default is False.
        :param <Tuple[float]> size: width and height of the figure
        :param <str> xlab: x-axis label. Default is no label.
        :param <str> ylab: y-axis label. Default is no label.
        :param <bool> fit_reg: if True (default) adds a regression line.
        :param <int> reg_ci: size of the confidence interval for the regression estimate.
            Default is 95. Can be None.
        :param <int> reg_order: order of the regression equation. Default is 1 (linear).
        :param <bool> reg_robust: if this will de-weight outliers, which is
            computationally expensive. Default is False.
        :param <str> coef_box: two-purpose parameter. Firstly, if the parameter equals
            None, the correlation coefficient and probability statistics will not be
            included in the graphic. Secondly, if the statistics are to be included,
            pass in the desired location on the graph which, in addition, will act as a
            pseudo-boolean parameter. The options are standard, i.e. 'upper left',
            'lower right' and so forth. Default is None, i.e the statistics are not
            displayed.
        :param <str> prob_est: type of estimator for probability of significant relation.
            The default is "pool", which means that all observation pairs of a panel
            are pooled and the probability is based on that pool.
            The alternative is "map", denoting Macrosynergy panel test. This is based
            on a panel regression with period-specific random effects and greatly
            mitigates the issue of pseudo-replication if panel features and targets
            are correlated across time.
            See also https://research.macrosynergy.com/testing-macro-trading-factors/
        :param <Union[str, int]> separator: allows categorizing the scatter analysis by
            cross-section or integer. In the former case the argument is set to
            "cids" and in the latter case the argument is set to a year [2010, for
            instance] which will subsequently split the time-period into the sample
            before (not including) that year and from (including) that year.
        :param <float> title_adj: parameter that sets top of figure to accommodate title.
            Default is 1.
        :param <bool> single_chart: boolean parameter determining whether the x- and y-
            labels are only written on a single graph of the Facet Grid (useful if there
            are numerous charts, and the labels are excessively long). The default is
            False, and the names of the axis will be displayed on each grid if not
            conflicting with the label for each variable.
        """

        coef_box_loc_error = "The parameter expects a string used to delimit the " \
                             "location of the box: 'upper left', 'lower right' etc."
        if coef_box is not None:
            assert isinstance(coef_box, str), coef_box_loc_error

        assert prob_est in ["pool", "map"], "prob_est must be 'pool' or 'map'"

        sns.set_theme(style="whitegrid")
        dfx = self.df.copy()

        if title is None and (self.years is None):
            dates = self.df.index.get_level_values('real_date').to_series().\
                dt.strftime('%Y-%m-%d')
            title = f'{self.xcats[0]} and {self.xcats[1]} ' \
                    f'from {dates.min()} to {dates.max()}'
        elif title is None:
            title = f'{self.xcats[0]} and {self.xcats[1]}'

        # If "separator" is type Integer, the scatter plot is split across two
        # time-periods where the divisor is the received year.
        if isinstance(separator, int):

            year_error = "Separation by years does not work with year groups."
            assert self.years is None, year_error

            fig, ax = plt.subplots(figsize=size)

            index_years = dfx.index.get_level_values(1).year
            years_in_df = list(index_years.unique())

            assert separator in years_in_df, "Separator year is not in the range."
            error_sep = "Separator year must not be the first in the range."
            assert separator > np.min(years_in_df), error_sep

            label_set1 = f"before {separator}"
            label_set2 = f"from {separator}"
            dfx1 = dfx[index_years < separator]
            dfx2 = dfx[index_years >= separator]

            sns.regplot(data=dfx1, x=self.xcats[0], y=self.xcats[1],
                        ci=reg_ci, order=reg_order, robust=reg_robust, fit_reg=fit_reg,
                        scatter_kws={'s': 30, 'alpha': 0.5},
                        label=label_set1,
                        line_kws={'lw': 1})
            sns.regplot(data=dfx2, x=self.xcats[0], y=self.xcats[1],
                        ci=reg_ci, order=reg_order, robust=reg_robust, fit_reg=fit_reg,
                        label=label_set2,
                        scatter_kws={'s': 30, 'alpha': 0.5},
                        line_kws={'lw': 1})

            if coef_box is not None:
                data_table = self.corr_probability(
                    df_probability=[dfx1, dfx2],
                    time_period="",
                    coef_box_loc=coef_box,
                    prob_est=prob_est
                )
                data_table.scale(0.4, 2.5)
                data_table.set_fontsize(14)

            ax.legend(loc='upper right')
            ax.set_title(title, fontsize=14)
            if xlab is not None:
                ax.set_xlabel(xlab)
            if ylab is not None:
                ax.set_ylabel(ylab)

        elif separator == "cids":

            assert isinstance(single_chart, bool)

            dfx_copy = dfx.reset_index()
            n_cids = len(dfx_copy['cid'].unique())

            error_cids = "There must be more than one cross-section to use " \
                         "separator = 'cids'."
            assert n_cids > 1, error_cids

            # “Wrap” the column variable at this width, so that the column facets span
            # multiple rows. Used to determine the number of grids on each row.
            dict_coln = {2: 2, 5: 3, 8: 4, 30: 5}

            keys_ar = np.array(list(dict_coln.keys()))
            key = keys_ar[keys_ar <= n_cids][-1]
            col_number = dict_coln[key]

            # The DataFrame is already a standardised DataFrame. Three columns: two
            # categories (dependent & explanatory variable) and the respective
            # cross-sections. The index will be the date timestamp.

            fg = sns.FacetGrid(data=dfx_copy, col='cid', col_wrap=col_number)
            fg.map(
                sns.regplot, self.xcats[0], self.xcats[1], ci=reg_ci, order=reg_order,
                robust=reg_robust, fit_reg=fit_reg,
                scatter_kws={'s': 15, 'alpha': 0.5, 'color': 'lightgray'},
                line_kws={'lw': 1}
            )

            if coef_box is not None:
                fg.map_dataframe(self.annotate_facet)

            fg.set_titles(col_template='{col_name}')
            fg.fig.suptitle(title, y=title_adj, fontsize=14)

            if not single_chart:
                if xlab is not None:
                    fg.set_xlabels(xlab, clear_inner=True)
                if ylab is not None:
                    fg.set_ylabels(ylab)
            else:
                error = "Label expected for the respective axis."
                assert xlab is not None, error
                assert ylab is not None, error
                number_of_graphs = len(fg.axes)
                no_columns = fg._ncol
                remainder = int(number_of_graphs % no_columns)

                for i in range(number_of_graphs):
                    fg.axes[i].set_xlabel("")
                    fg.axes[i].set_ylabel("")

                    if remainder == 0:
                        fg.axes[no_columns].set_xlabel(xlab)
                        fg.axes[no_columns].set_ylabel(ylab)
                    else:
                        fg.axes[-remainder].set_xlabel(xlab)
                        fg.axes[-remainder].set_ylabel(ylab)

        elif separator is None:
            fig, ax = plt.subplots(figsize=size)

            sns.regplot(data=dfx, x=self.xcats[0], y=self.xcats[1],
                        ci=reg_ci, order=reg_order, robust=reg_robust, fit_reg=fit_reg,
                        scatter_kws={'s': 30, 'alpha': 0.5, 'color': 'lightgray'},
                        line_kws={'lw': 1})

            if coef_box is not None:
                data_table = self.corr_probability(
                    df_probability=self.df, prob_est=prob_est,
                    coef_box_loc=coef_box
                )
                data_table.scale(0.4, 2.5)
                data_table.set_fontsize(12)

            if labels:
                error_freq = "Labels only available for monthly or lower frequencies."
                assert self.freq in ['A', 'Q', 'M'], error_freq

                df_labs = self.df.dropna().index.to_frame(index=False)
                if self.years is not None:
                    ser_labs = df_labs['cid'] + ' ' + df_labs['real_date']
                else:
                    ser_labs = df_labs['cid'] + ' '
                    ser_labs += df_labs['real_date'].dt.year.astype(str)
                    if self.freq == 'Q':
                        ser_labs += 'Q' + df_labs['real_date'].dt.quarter.astype(str)

                    elif self.freq == 'M':
                        ser_labs += '-' + df_labs['real_date'].dt.month.astype(str)

                for i in range(self.df.shape[0]):

                    plt.text(x=self.df[self.xcats[0]][i] + 0,
                             y=self.df[self.xcats[1]][i] + 0, s=ser_labs[i],
                             fontdict=dict(color='black', size=8))

            ax.set_title(title, fontsize=14)
            if xlab is not None:
                ax.set_xlabel(xlab)
            if ylab is not None:
                ax.set_ylabel(ylab)

        else:
            ValueError("Separator must be either a valid year <int> or 'cids' <str>.")
            
        plt.show()

    def ols_table(self, type='pool'):
        """
        Print statsmodels regression summaries.
        :param <str> type: type of linear regression summary to print. Default is 'pool'.
            Alternative is 're' for period-specific random effects.

        """
        assert type in ['pool', 're'], "Type must be either 'pool' or 're'."

        x, y = self.df.dropna().iloc[:, 0], self.df.dropna().iloc[:, 1]
        x_fit = sm.add_constant(x)
        groups = self.df.reset_index().real_date
        if type == 'pool':
            fit_results = sm.OLS(y, x_fit).fit()
        elif type == 're':
            fit_results = sm.MixedLM(y, x_fit, groups).fit(reml=False)
        
        print(fit_results.summary())


if __name__ == "__main__":

    cids = ['AUD', 'CAD', 'GBP', 'NZD', 'USD']
    xcats = ['XR', 'CRY', 'GROWTH', 'INFL']
    df_cids = pd.DataFrame(index=cids,
                           columns=['earliest', 'latest', 'mean_add', 'sd_mult'])
    df_cids.loc['AUD'] = ['2000-01-01', '2020-12-31', 0.1, 1]
    df_cids.loc['CAD'] = ['2001-01-01', '2020-11-30', 0, 1]
    df_cids.loc['BRL'] = ['2001-01-01', '2020-11-30', -0.1, 2]
    df_cids.loc['GBP'] = ['2002-01-01', '2020-11-30', 0, 2]
    df_cids.loc['NZD'] = ['2002-01-01', '2020-09-30', -0.1, 2]
    df_cids.loc['USD'] = ['2003-01-01', '2020-12-31', -0.1, 2]

    cols = ['earliest', 'latest', 'mean_add', 'sd_mult', 'ar_coef', 'back_coef']
    df_xcats = pd.DataFrame(index=xcats, columns=cols)
    df_xcats.loc['XR'] = ['2000-01-01', '2020-12-31', 0.1, 1, 0, 0.3]
    df_xcats.loc['CRY'] = ['2000-01-01', '2020-10-30', 1, 2, 0.95, 1]
    df_xcats.loc['GROWTH'] = ['2001-01-01', '2020-10-30', 1, 2, 0.9, 1]
    df_xcats.loc['INFL'] = ['2001-01-01', '2020-10-30', 1, 2, 0.8, 0.5]

    dfd = make_qdf(df_cids, df_xcats, back_ar=0.75)
    dfd["grading"] = np.ones(dfd.shape[0])
    black = {'AUD': ['2000-01-01', '2003-12-31'], 'GBP': ['2018-01-01', '2100-01-01']}

    # All AUD GROWTH locations.
    filt1 = (dfd['xcat'] == 'GROWTH') & (dfd['cid'] == 'AUD')
    filt2 = (dfd['xcat'] == 'INFL') & (dfd['cid'] == 'NZD')

    # Reduced DataFrame.
    dfdx = dfd[~(filt1 | filt2)]
    dfdx['ERA'] = "before 2010"
    dfdx.loc[dfdx.real_date.dt.year > 2007, 'ERA'] = "from 2010"

    cidx = ['AUD', 'CAD', 'GBP', 'USD']

    cr = CategoryRelations(
        dfdx, xcats=["CRY", "XR"], freq="M", lag=1, cids=cidx, xcat_aggs=["mean", "sum"],
        start="2001-01-01", blacklist=black, years=None
    )

    cr.reg_scatter(
        labels=False, separator=None, title="Carry and Return", xlab="Carry",
        ylab="Return", coef_box="lower left", prob_est="map",
    )

    # years parameter

    cr = CategoryRelations(
        dfdx, xcats=["CRY", "XR"], freq="M", years=5, lag=0, cids=cidx, xcat_aggs=["mean", "sum"],
        start="2001-01-01", blacklist=black
    )

    cr.reg_scatter(
        labels=False, separator=None, title="Carry and Return, 5-year periods", xlab="Carry",
        ylab="Return", coef_box="lower left", prob_est="map",
    )

    cr = CategoryRelations(
        dfdx, xcats=["CRY", "XR"], 
        xcat1_chg="diff",
        freq="M", 
        lag=1, 
        cids=cidx, 
        xcat_aggs=["mean", "sum"],
        start="2001-01-01", 
        blacklist=black, years=None
    )

    cr.reg_scatter(
        labels=False, separator=cids, title="Carry and Return", xlab="Carry",
        ylab="Return", coef_box="lower left"
    )

    cr.ols_table(type='pool')
    cr.ols_table(type='re')


import pandas as pd
import seaborn as sns
import matplotlib.pyplot as plt

from typing import List, Tuple, Optional
from macrosynergy.management.simulate_quantamental_data import make_test_df
from macrosynergy.management.utils import is_valid_iso_date
from macrosynergy.management.shape_dfs import reduce_df


def view_metrics(
    df: pd.DataFrame,
    xcat: str,
    cids: Optional[List[str]] = None,
    start: Optional[str] = None,
    end: Optional[str] = None,
    freq: str = "M",
    agg: str = "mean",
    metric: str = "eop_lag",
    title: Optional[str] = None,
    figsize: Optional[Tuple[float]] = (14, None),
) -> None:
    """
    A function to visualise the `eop_lag`, `mop_lag` or `grading` metrics for a given
    JPMaQS dataset. It generates a heatmap, where the x-axis is the observation
    date, the y-axis is the ticker, and the colour is the lag value.

    :param <pd.Dataframe> df: standardized DataFrame with the necessary columns:
        'cid', 'xcat', 'real_date' and 'grading', 'eop_lag' or 'mop_lag'.
    :param str xcat: extended category whose lags are to be visualized.
    :param <List[str]> cids: cross sections to visualize. Default is all in DataFrame.
    :param <str> start: earliest date in ISO format. Default is earliest available.
    :param <str> end: latest date in ISO format. Default is latest available.
    :param <str> metric: name of metric to be visualized. Must be "eop_lag" (default)
        "mop_lag" or "grading".
    :param <str> freq: frequency of data. Must be one of "D", "W", "M", "Q", "A".
        Default is "M".
    :param <str> agg: aggregation method. Must be one of "mean" (default), "median",
        "min", "max", "first" or "last".
    :param <str> title: string of chart title; if none given default title is printed.
    :param <Tuple[float]> figsize: Tuple (w, h) of width and height of graph.
        Default is None, meaning it is set in accordance with df.

    :return: None

    :raises TypeError: if any of the inputs are of the wrong type.
    :raises ValueError: if any of the inputs are semantically incorrect.
    """
    # Validating inputs
    # First check if the standard valid and expected columns are present
    if not isinstance(metric, str):
        raise TypeError("`metric` must be a string")
    elif metric not in ["eop_lag", "mop_lag", "grading"]:
        raise ValueError("`metric` must be either 'eop_lag', 'mop_lag' or 'grading'")

    expc_cols: List[str] = ["cid", "xcat", "real_date", metric]
    if not isinstance(df, pd.DataFrame):
        raise TypeError("`df` must be a DataFrame")
    elif df.empty:
        raise ValueError("`df` must not be empty")
    elif not set(expc_cols).issubset(df.columns):
        raise ValueError(
            f"`df` must have columns 'cid', 'xcat', 'real_date' and '{metric}'"
        )
    else:
        # Filter to only necessary columns
        df = df.copy()[expc_cols]

    # Check if xcat is a single string and is in the DataFrame

    if not isinstance(xcat, str):
        raise TypeError("`xcat` must be a string")
    else:
        if not xcat in df["xcat"].unique():
            raise ValueError(f"No data for `xcat` = {xcat} in DataFrame `df`")

    # Check if cids is a list of strings and is in the DataFrame
    if cids is not None:
        if (
            (not isinstance(cids, list))
            or (not all(isinstance(cid, str) for cid in cids))
            or (len(cids) == 0)
        ):
            raise TypeError("`cids` must be a list of strings")

        if not set(cids).issubset(df["cid"].unique()):
            raise ValueError("All cids in `cids` must be in the DataFrame `df`")

    else:
        # In case cids is None, set it to all cids in the DataFrame
        cids = list(df["cid"].unique())

    # Validate start and end dates
    df["real_date"]: pd.DatetimeIndex = pd.to_datetime(df["real_date"])
    min_date: pd.Timestamp = df["real_date"].min()
    max_date: pd.Timestamp = df["real_date"].max()
    if start is None:
        start: str = min_date.strftime("%Y-%m-%d")
    if end is None:
        end: str = max_date.strftime("%Y-%m-%d")

    for varx, namex in zip([start, end], ["start", "end"]):
        if not is_valid_iso_date(varx):
            raise ValueError(f"`{namex}` must be a valid ISO date string")

    # Validate freq and agg
    if not isinstance(freq, str):
        raise TypeError("`freq` must be a string")
    else:
        freq: str = freq.upper()
        if freq not in ["D", "W", "M", "Q", "A"]:
            raise ValueError("`freq` must be one of 'D', 'W', 'M', 'Q' or 'A'")

    if not isinstance(agg, str):
        raise TypeError("`agg` must be a string")
    else:
        agg: str = agg.lower()
        if agg not in ["mean", "median", "min", "max", "first", "last"]:
            raise ValueError(
                "`agg` must be one of 'mean', 'median', 'min', 'max', 'first' or 'last'"
            )

    # Validate title and figsize tuple.
    # Figsize tuple allows second value to be None,
    # which is then set to the number of cids
    if title is not None:
        if not isinstance(title, str):
            raise TypeError("`title` must be a string")

    if not isinstance(figsize, tuple):
        if not isinstance(figsize, list) or len(figsize) != 2:
            raise TypeError("`figsize` must be a tuple or list of length 2")

    if not isinstance(figsize[0], (int, float)):
        raise ValueError("First element of `figsize` must be a float")

    if figsize[1] is None:
        figsize = (figsize[0], len(cids) + 1) # +1 to adjust for labels

    if not isinstance(figsize[1], (int, float)):
        raise ValueError("Second element of `figsize` must be a float")

    # Actual plotting code

    # filter and reduce the df to the required cids, xcat, start and end
    filtered_df: pd.DataFrame = reduce_df(
        df=df, cids=cids, xcats=[xcat], start=start, end=end
    )
    # group by cid, then set the index to real_date, then resample to the required freq and agg
    filtered_df: pd.DataFrame = (
        filtered_df.groupby(["cid", "xcat"])
        .apply(
            lambda x: x.set_index("real_date")
            .resample(freq)
            .agg(agg, numeric_only=True)
        )
        .reset_index()
    )

    filtered_df["real_date"]: pd.Series = filtered_df["real_date"].dt.strftime(
        "%Y-%m-%d"
    )

    pivoted_df: pd.DataFrame = filtered_df.pivot_table(
        index="cid", columns="real_date", values=metric
    )

    fig, ax = plt.subplots(figsize=figsize, layout="constrained")
    max_mes: float = max(1, filtered_df[metric].max())
    min_mes: float = min(0, filtered_df[metric].min())
    sns.heatmap(
        pivoted_df,
        cmap=sns.color_palette("light:red", as_cmap=True),
        vmin=min_mes,
        vmax=max_mes,
        ax=ax,
    )

    if title is None:
        title = f"Visualising {metric} for {xcat} from {start} to {end}"
    ax.set_title(title)
    ax.set_xlabel("Date")
    ax.set_ylabel("Cross Sections")
    plt.show()


if __name__ == "__main__":
    test_cids: List[str] = ["USD",]#"EUR", "GBP"]
    test_xcats: List[str] = ["FX", "IR"]
    dfE: pd.DataFrame = make_test_df(
        cids=test_cids, xcats=test_xcats, prefer="sharp-hill"
    )

    dfM: pd.DataFrame = make_test_df(
        cids=test_cids, xcats=test_xcats, prefer="four-bit-sine"
    )

    dfG: pd.DataFrame = make_test_df(cids=test_cids, xcats=test_xcats, prefer="sine")

    dfE.rename(columns={"value": "eop_lag"}, inplace=True)
    dfM.rename(columns={"value": "mop_lag"}, inplace=True)
    dfG.rename(columns={"value": "grading"}, inplace=True)
    mergeon = ["cid", "xcat", "real_date"]
    dfx: pd.DataFrame = pd.merge(pd.merge(dfE, dfM, on=mergeon), dfG, on=mergeon)

    view_metrics(
        df=dfx,
        xcat="FX",
    )
    view_metrics(
        df=dfx,
        xcat="IR",
        metric="mop_lag",
    )
    view_metrics(
        df=dfx,
        xcat="IR",
        metric="grading",
    )




import numpy as np
import pandas as pd
from typing import List, Dict, Set
import warnings
from macrosynergy.management.shape_dfs import reduce_df
from macrosynergy.management.simulate_quantamental_data import make_qdf
from macrosynergy.management.utils import drop_nan_series

def expanding_stat(df: pd.DataFrame, dates_iter: pd.DatetimeIndex,
                   stat: str = 'mean', sequential: bool = True,
                   min_obs: int = 261, iis: bool = True):

    """
    Compute statistic based on an expanding sample.

    :param <pd.Dataframe> df: Daily-frequency time series DataFrame.
    :param <pd.DatetimeIndex> dates_iter: controls the frequency of the neutral &
        standard deviation calculations.
    :param <str> stat: statistical method to be applied. This is typically 'mean',
        or 'median'.
    :param <bool> sequential: if True (default) the statistic is estimated sequentially.
        If this set to false a single value is calculated per time series, based on
        the full sample.
    :param <int> min_obs: minimum required observations for calculation of the
        statistic in days.
    :param <bool> iis: if set to True, the values of the initial interval determined
        by min_obs will be estimated in-sample, based on the full initial sample.

    :return: Time series dataframe of the chosen statistic across all columns
    """

    df_out = pd.DataFrame(np.nan, index=df.index, columns=['value'])
    # An adjustment for individual series' first realised value is not required given the
    # returned DataFrame will be subtracted from the original DataFrame. The original
    # DataFrame will implicitly host this information through NaN values such that when
    # the arithmetic operation is made, any falsified values will be displaced by NaN
    # values.

    first_observation = df.dropna(axis=0, how='all').index[0]
    # Adjust for individual cross-sections' series commencing at different dates.
    first_estimation = df.dropna(axis=0, how='all').index[min_obs]

    obs_index = next(iter(np.where(df.index == first_observation)))[0]
    est_index = next(iter(np.where(df.index == first_estimation)))[0]

    if stat == "zero":

        df_out["value"] = 0

    elif not sequential:
        # The entire series is treated as in-sample. Will automatically handle NaN
        # values.
        statval = df.stack().apply(stat)
        df_out["value"] = statval

    else:

        dates = dates_iter[dates_iter >= first_estimation]
        for date in dates:

            df_out.loc[date, "value"] = df.loc[first_observation:date].stack().apply(stat)

        df_out = df_out.fillna(method='ffill')

        if iis:
            if (est_index - obs_index)>0:
                df_out = df_out.fillna(method="bfill", limit=(est_index - obs_index))

    df_out.columns.name = 'cid'
    return df_out

def make_zn_scores(df: pd.DataFrame, xcat: str, cids: List[str] = None,
                   start: str = None, end: str = None, blacklist: dict = None,
                   sequential: bool = True, min_obs: int = 261, iis: bool = True,
                   neutral: str = 'zero', est_freq: str = 'd', thresh: float = None,
                   pan_weight: float = 1, postfix: str = 'ZN'):

    """
    Computes z-scores for a panel around a neutral level ("zn scores").
    
    :param <pd.Dataframe> df: standardized JPMaQS DataFrame with the necessary columns:
        'cid', 'xcat', 'real_date' and 'value'.
    :param <str> xcat:  extended category for which the zn_score is calculated.
    :param <List[str]> cids: cross sections for which zn_scores are calculated; default
        is all available for category.
    :param <str> start: earliest date in ISO format. Default is None and earliest date in
        df is used.
    :param <str> end: latest date in ISO format. Default is None and latest date in df is
        used.
    :param <dict> blacklist: cross-sections with date ranges that should be excluded from
        the calculation of zn-scores.
        This means that not only are there no zn-score values calculated for these
        periods, but also that they are not used for the scoring of other periods.
        N.B.: The argument is a dictionary with cross-sections as keys and tuples of
        start and end dates of the blacklist periods in ISO formats as values.
        If one cross section has multiple blacklist periods, numbers are added to the
        keys (i.e. TRY_1, TRY_2, etc.)
    :param <bool> sequential: if True (default) score parameters (neutral level and
        standard deviations) are estimated sequentially with concurrently available
        information only.
    :param <int> min_obs: the minimum number of observations required to calculate
        zn_scores. Default is 261. The parameter is only applicable if the "sequential"
        parameter is set to True. Otherwise the neutral level and the standard deviation
        are both computed in-sample and will use the full sample.
    :param <bool> iis: if True (default) zn-scores are also calculated for the initial
        sample period defined by min-obs on an in-sample basis to avoid losing history.
        This is irrelevant if sequential is set to False.
    :param <str> neutral: method to determine neutral level. Default is 'zero'.
        Alternatives are 'mean' and "median".
    :param <str> est_freq: the frequency at which standard deviations or means are
        are re-estimated. The options are weekly, monthly & quarterly "w", "m", "q".
        Default is daily, "d". Re-estimation is performed at period end.
    :param <float> thresh: threshold value beyond which scores are winsorized,
        i.e. contained at that threshold. The threshold is the maximum absolute
        score value that the function is allowed to produce. The minimum threshold is 1
        standard deviation.
    :param <float> pan_weight: weight of panel (versus individual cross section) for
        calculating the z-score parameters, i.e. the neutral level and the standard
        deviation. Default is 1, i.e. panel data are the basis for the parameters.
        Lowest possible value is 0, i.e. parameters are all specific to cross section.
    :param <str> postfix: string appended to category name for output; default is "ZN".

    :return <pd.Dataframe>: standardized DataFrame with the zn-scores of the chosen xcat:
        'cid', 'xcat', 'real_date' and 'value'.
    """
    df = df.copy()
    df["real_date"] = pd.to_datetime(df["real_date"], format="%Y-%m-%d")

    expected_columns = ["cid", "xcat", "real_date", "value"]
    col_error = f"The DataFrame must contain the necessary columns: {expected_columns}."
    assert set(expected_columns).issubset(set(df.columns)), col_error

    # --- Assertions

    assert neutral in ["mean", "median", "zero"]
    if thresh is not None:
        assert thresh > 1, "The 'thresh' parameter must be larger than 1."
    assert 0 <= pan_weight <= 1, "The 'pan_weight' parameter must be between 0 and 1."
    assert isinstance(iis, bool), "Boolean Object required."

    error_min = "Minimum observations must be a non-negative Integer value."
    assert isinstance(min_obs, int) and min_obs >= 0, error_min

    frequencies = ["d", "w", "m", "q"]
    error_freq = f"String Object required and must be one of the available frequencies: " \
                 f"{frequencies}."
    assert isinstance(est_freq, str) and est_freq in frequencies, error_freq
    pd_freq = dict(zip(frequencies, ['B', 'W-Fri', 'BM', 'BQ']))

    # --- Prepare re-estimation dates and time-series DataFrame.

    # Remove any additional metrics defined in the DataFrame.
    df = df.loc[:, expected_columns]
    if cids is not None:
        missing_cids = set(cids).difference(set(df['cid']))
        if missing_cids:
            raise ValueError(f"The following cids are not available in the DataFrame: "
                                f"{missing_cids}.")
    if xcat not in df['xcat'].unique():
        raise ValueError(f"The xcat {xcat} is not available in the DataFrame.")
    
    df = reduce_df(
        df, xcats=[xcat], cids=cids, start=start, end=end, blacklist=blacklist
    )

    if df.isna().values.any():
        df = drop_nan_series(df=df, raise_warning=True)

    s_date = min(df['real_date'])
    e_date = max(df['real_date'])
    dates_iter = pd.date_range(start=s_date, end=e_date, freq=pd_freq[est_freq])

    dfw = df.pivot(index='real_date', columns='cid', values='value')
    cross_sections = dfw.columns

    # --- The actual scoring.

    dfw_zns_pan = dfw * 0
    dfw_zns_css = dfw * 0

    if pan_weight > 0:

        df_neutral = expanding_stat(dfw, dates_iter, stat=neutral,
                                    sequential=sequential, min_obs=min_obs,
                                    iis=iis)
        dfx = dfw.sub(df_neutral['value'], axis=0)
        df_mabs = expanding_stat(dfx.abs(), dates_iter, stat="mean",
                                 sequential=sequential,
                                 min_obs=min_obs, iis=iis)
        dfw_zns_pan = dfx.div(df_mabs['value'], axis='rows')

    if pan_weight < 1:

        for cid in cross_sections:
            dfi = dfw[cid]

            df_neutral = expanding_stat(dfi.to_frame(name=cid), dates_iter, stat=neutral,
                                        sequential=sequential,
                                        min_obs=min_obs, iis=iis)
            dfx = dfi - df_neutral['value']

            df_mabs = expanding_stat(dfx.abs().to_frame(name=cid), dates_iter,
                                     stat="mean", sequential=sequential,
                                     min_obs=min_obs, iis=iis)
            dfx = pd.DataFrame(data=dfx.to_numpy(),
                               index=dfx.index, columns=['value'])
            dfx = dfx.rename_axis('cid', axis=1)

            zns_css_df = dfx / df_mabs
            dfw_zns_css.loc[:, cid] = zns_css_df.to_numpy()

    dfw_zns = (dfw_zns_pan * pan_weight) + (dfw_zns_css * (1 - pan_weight))
    dfw_zns = dfw_zns.dropna(axis=0, how='all')
    
    if thresh is not None:
        dfw_zns.clip(lower=-thresh, upper=thresh, inplace=True)

    # --- Reformatting of output into standardised DataFrame.

    df_out = dfw_zns.stack().to_frame("value").reset_index()
    df_out['xcat'] = xcat + postfix

    col_names = ['cid', 'xcat', 'real_date', 'value']
    df_out = df_out.sort_values(['cid', 'real_date'])[col_names]

    return df_out[df.columns].reset_index(drop=True)


if __name__ == "__main__":


    cids = ['AUD', 'CAD', 'GBP', 'USD', 'NZD']
    xcats = ['XR', 'CRY', 'GROWTH', 'INFL']

    df_cids = pd.DataFrame(index = cids, columns = ['earliest', 'latest', 'mean_add',
                                                    'sd_mult'])

    df_cids.loc['AUD'] = ['2010-01-01', '2020-12-31', 0.5, 2]
    df_cids.loc['CAD'] = ['2006-01-01', '2020-11-30', 0, 1]
    df_cids.loc['GBP'] = ['2008-01-01', '2020-11-30', -0.2, 0.5]
    df_cids.loc['USD'] = ['2007-01-01', '2020-09-30', -0.2, 0.5]
    df_cids.loc['NZD'] = ['2002-01-01', '2020-09-30', -0.1, 2]

    df_xcats = pd.DataFrame(index = xcats, columns = ['earliest', 'latest', 'mean_add',
                                                      'sd_mult', 'ar_coef', 'back_coef'])
    df_xcats.loc['XR'] = ['2008-01-01', '2020-12-31', 0, 1, 0, 0.3]
    df_xcats.loc['CRY'] = ['2011-01-01', '2020-10-30', 1, 2, 0.9, 0.5]
    df_xcats.loc['GROWTH'] = ['2012-01-01', '2020-10-30', 1, 2, 0.9, 1]
    df_xcats.loc['INFL'] = ['2013-01-01', '2020-10-30', 1, 2, 0.8, 0.5]

    # Apply a blacklist period from series' start date.
    black = {'AUD': ['2010-01-01', '2013-12-31'],
             'GBP': ['2018-01-01', '2100-01-01']}

    dfd = make_qdf(df_cids, df_xcats, back_ar = 0.75)
    dfd["grading"] = np.ones(dfd.shape[0])

    # Monthly: panel + cross.
    dfzm = make_zn_scores(dfd, xcat='XR', sequential=True, cids=cids,
                          blacklist=black, iis=True, neutral='mean',
                          pan_weight=0.75, min_obs=261, est_freq="m")
    print(dfzm)

    # Weekly: panel + cross.
    dfzw = make_zn_scores(dfd, xcat='XR', sequential=True, cids=cids,
                          blacklist=black, iis=False, neutral='mean',
                          pan_weight=0.5, min_obs=261, est_freq="w")

    # Daily: panel. Neutral and standard deviation will be computed daily.
    dfzd = make_zn_scores(dfd, xcat='XR', sequential=True, cids=cids,
                          blacklist=black, iis=True, neutral='mean',
                          pan_weight=1.0, min_obs=261, est_freq="d")

    # Daily: cross.
    dfd['ticker'] = dfd["cid"] + "_" + dfd["xcat"]
    dfzd = make_zn_scores(dfd, xcat='XR', sequential=True, cids=cids,
                          blacklist=black, iis=True, neutral='mean',
                          pan_weight=0.0, min_obs=261, est_freq="d")

    panel_df = make_zn_scores(dfd, 'CRY', cids, start="2010-01-04", blacklist=black,
                              sequential=False, min_obs=0, neutral='mean',
                              iis=True, thresh=None, pan_weight=0.75, postfix='ZN')



import numpy as np
import pandas as pd
from statsmodels.tsa.arima_process import ArmaProcess
from typing import List, Tuple, Dict, Union
from collections import defaultdict
import datetime


def simulate_ar(nobs: int, mean: float = 0, sd_mult: float = 1, ar_coef: float = 0.75):
    """
    Create an auto-correlated data-series as numpy array.

    :param <int> nobs: number of observations.
    :param <float> mean: mean of values, default is zero.
    :param <float> sd_mult: standard deviation multipliers of values, default is 1.
        This affects non-zero means.
    :param <float> ar_coef: autoregression coefficient (between 0 and 1): default is 0.75.

    return <np.array>: autocorrelated data series.
    """

    # Define relative parameters for creating an AR process.
    ar_params = np.r_[1, -ar_coef]
    ar_proc = ArmaProcess(ar_params)  # define ARMA process
    ser = ar_proc.generate_sample(nobs)
    ser = ser + mean - np.mean(ser)
    return sd_mult * ser / np.std(ser)

def dataframe_generator(df_cids: pd.DataFrame, df_xcats: pd.DataFrame,
                        cid: str, xcat: str):
    """
    Adjacent method used to construct the quantamental DataFrame.

    :param <pd.DataFrame> df_cids:
    :param <pd.DataFrame> df_xcats:
    :param <str> cid: individual cross-section.
    :param <str> xcat: individual category.

    """
    qdf_cols = ['cid', 'xcat', 'real_date', 'value']

    sdate = pd.to_datetime(max(df_cids.loc[cid, 'earliest'], df_xcats.loc[xcat,
                                                                          'earliest']))
    edate = pd.to_datetime(min(df_cids.loc[cid, 'latest'], df_xcats.loc[xcat,
                                                                        'latest']))
    all_days = pd.date_range(sdate, edate)
    work_days = all_days[all_days.weekday < 5]

    df_add = pd.DataFrame(columns=qdf_cols)
    df_add['real_date'] = work_days
    df_add['cid'] = cid
    df_add['xcat'] = xcat

    return df_add, work_days

def make_qdf(df_cids: pd.DataFrame, df_xcats: pd.DataFrame, back_ar: float = 0):
    """
    Make quantamental DataFrame with basic columns: 'cid', 'xcat', 'real_date', 'value'.

    :param <pd.DataFrame> df_cids: DataFrame with parameters by cid. Row indices are
        cross-sections. Columns are:
        'earliest': string of earliest date (ISO) for which country values are available;
        'latest': string of latest date (ISO) for which country values are available;
        'mean_add': float of country-specific addition to any category's mean;
        'sd_mult': float of country-specific multiplier of an category's standard
            deviation.
    :param <pd.DataFrame> df_xcats: dataframe with parameters by xcat. Row indices are
        cross-sections. Columns are:
        'earliest': string of earliest date (ISO) for which category values are
        available;
        'latest': string of latest date (ISO) for which category values are available;
        'mean_add': float of category-specific addition;
        'sd_mult': float of country-specific multiplier of an category's standard
        deviation;
        'ar_coef': float between 0 and 1 denoting set auto-correlation of the category;
        'back_coef': float, coefficient with which communal (mean 0, SD 1) background
        factor is added to category values.
    :param <float> back_ar: float between 0 and 1 denoting set auto-correlation of the
        background factor. Default is zero.

    :return <pd.DataFrame>: basic quantamental DataFrame according to specifications.
    """
    df_list = []

    if any(df_xcats['back_coef'] != 0):

        sdate = min(min(df_cids.loc[:, 'earliest']), min(df_xcats.loc[:, 'earliest']))
        edate = max(max(df_cids.loc[:, 'latest']), max(df_xcats.loc[:, 'latest']))
        all_days = pd.date_range(sdate, edate)
        work_days = all_days[all_days.weekday < 5]
        ser = simulate_ar(len(work_days), mean=0, sd_mult=1, ar_coef=back_ar)
        df_back = pd.DataFrame(index=work_days, columns=['value'])
        df_back['value'] = ser

    for cid in df_cids.index:
        for xcat in df_xcats.index:
            df_add, work_days = dataframe_generator(df_cids=df_cids, df_xcats=df_xcats,
                                                    cid=cid, xcat=xcat)

            ser_mean = df_cids.loc[cid, 'mean_add'] + df_xcats.loc[xcat, 'mean_add']
            ser_sd = df_cids.loc[cid, 'sd_mult'] * df_xcats.loc[xcat, 'sd_mult']
            ser_arc = df_xcats.loc[xcat, 'ar_coef']
            df_add['value'] = simulate_ar(len(work_days), mean=ser_mean, sd_mult=ser_sd,
                                          ar_coef=ser_arc)

            back_coef = df_xcats.loc[xcat, 'back_coef']
            # Add the influence of communal background series.
            if back_coef != 0:
                dates = df_add['real_date']
                df_add['value'] = df_add['value'] + \
                                  back_coef * df_back.loc[dates,
                                                          'value'].reset_index(drop=True)

            df_list.append(df_add)

    return pd.concat(df_list).reset_index(drop=True)

def make_qdf_black(df_cids: pd.DataFrame, df_xcats: pd.DataFrame, blackout: dict):
    """
    Make quantamental DataFrame with basic columns: 'cid', 'xcat', 'real_date', 'value'.
    In this DataFrame the column, 'value', will consist of Binary Values denoting whether
    the cross-section is active for the corresponding dates.

    :param <pd.DataFrame> df_cids: dataframe with parameters by cid. Row indices are
        cross-sections. Columns are:
    'earliest': string of earliest date (ISO) for which country values are available;
    'latest': string of latest date (ISO) for which country values are available;
    'mean_add': float of country-specific addition to any category's mean;
    'sd_mult': float of country-specific multiplier of an category's standard deviation.
    :param <pd.DataFrame> df_xcats: dataframe with parameters by xcat. Row indices are
        cross-sections. Columns are:
    'earliest': string of earliest date (ISO) for which category values are available;
    'latest': string of latest date (ISO) for which category values are available;
    'mean_add': float of category-specific addition;
    'sd_mult': float of country-specific multiplier of an category's standard deviation;
    'ar_coef': float between 0 and 1 denoting set autocorrelation of the category;
    'back_coef': float, coefficient with which communal (mean 0, SD 1) background
        factor is added to categoy values.
    :param <dict> blackout: Dictionary defining the blackout periods for each cross-
        section. The expected form of the dictionary is:
        {'AUD': (Timestamp('2000-01-13 00:00:00'), Timestamp('2000-01-13 00:00:00')),
        'USD_1': (Timestamp('2000-01-03 00:00:00'), Timestamp('2000-01-05 00:00:00')),
        'USD_2': (Timestamp('2000-01-09 00:00:00'), Timestamp('2000-01-10 00:00:00')),
        'USD_3': (Timestamp('2000-01-12 00:00:00'), Timestamp('2000-01-12 00:00:00'))}
        The values of the dictionary are tuples consisting of the start & end-date of the
        respective blackout period. Each cross-section could have potentially more than
        one blackout period on a single category, and subsequently each key will be
        indexed to indicate the number of periods.

    :return <pd.DataFrame>: basic quantamental DataFrame according to specifications with
        binary values.
    """

    df_list = []

    conversion = lambda t: (pd.Timestamp(t[0]), pd.Timestamp(t[1]))
    dates_dict = defaultdict(list)
    for k, v in blackout.items():
        v = conversion(v)
        dates_dict[k[:3]].append(v)

    # At the moment the blackout period is being applied uniformally to each category:
    # each category will experience the same blackout periods.
    for cid in df_cids.index:
        for xcat in df_xcats.index:

            df_add, work_days = dataframe_generator(df_cids=df_cids, df_xcats=df_xcats,
                                                    cid=cid, xcat=xcat)
            arr = np.repeat(0, df_add.shape[0])
            dates = df_add['real_date'].to_numpy()

            list_tuple = dates_dict[cid]
            for tup in list_tuple:

                start = tup[0]
                end = tup[1]

                condition_start = start.weekday() - 4
                condition_end = end.weekday() - 4

                # Will skip the associated blackout period because of the received
                # invalid date, if it is not within the respective data series' range,
                # but will continue to populate the dataframe according to the other keys
                # in the dictionary.
                # Naturally compare against the data-series' formal start & end date.
                if start < dates[0] or end > dates[-1]:
                    print("Blackout period date not within data series range.")
                    break
                # If the date falls on a weekend, change to the following Monday.
                elif condition_start > 0:
                    while start.weekday() > 4:
                        start += datetime.timedelta(days=1)
                elif condition_end > 0:
                    while end.weekday() > 4:
                        end += datetime.timedelta(days=1)

                index_start = next(iter(np.where(dates == start)[0]))
                count = 0
                while start != tup[1]:
                    if start.weekday() < 5:
                        count += 1
                    start += datetime.timedelta(days=1)

                arr[index_start:(index_start + count + 1)] = 1

            df_add['value'] = arr

            df_list.append(df_add)

    return pd.concat(df_list).reset_index(drop=True)


def generate_lines(sig_len: int,
                   style : str = 'linear') -> Union[np.ndarray,
                                                    Dict[str, np.ndarray]]:
    
    if not isinstance(sig_len, int):
        raise TypeError("`sig_len` must be an integer.")
    elif sig_len < 1:
        raise ValueError("`sig_len` must be greater than 0.")
    
    if not isinstance(style, str):
        raise TypeError("`style` must be a string.")
    
    style : str = '-'.join(style.strip().lower().split())
    lines: Dict[str, np.ndarray] = {
            "linear": np.arange(1, sig_len + 1) * 100 / sig_len,
            "decreasing-linear": np.arange(sig_len, 0, -1) * 100 / sig_len,
            "sharp-hill": np.concatenate(
                (
                    np.arange(1, sig_len // 4 + 1) * 100 / sig_len,
                    np.arange(sig_len // 4, sig_len // 4 * 3 + 1)[::-1] * 100 / sig_len,
                    np.arange(sig_len // 4 * 3, sig_len + 1) * 100 / sig_len,
                )
            ),
            "four-bit-sine": np.concatenate(
                (
                    np.arange(1, sig_len // 4 + 1) * 100 / sig_len,
                    np.arange(sig_len // 4, sig_len // 4 * 3 + 1)[::-1] * 100 / sig_len,
                    np.arange(sig_len // 4 * 3, sig_len + 1) * 100 / sig_len,
                )
            ),
            # sine wave with 1==sig_len peak=50, trough=-50
            "sine": np.sin(np.arange(1, sig_len + 1) * np.pi / (sig_len / 2)) * 50 + 50,
        }
    for k, v in lines.items():
        lines[k] = v[:sig_len]
    if style == "any":
        return list(lines.values())[np.random.randint(0, len(lines))]
    elif style == "all":
        return lines
    elif style in lines:
        return lines[style]
    else:
        raise ValueError(f"Invalid style: {style}. Use one of: {list(lines.keys())}.")
    


def make_test_df(
    cids: List[str] = ["AUD", "CAD", "GBP"],
    xcats: List[str] = ["XR", "CRY"],
    start_date: str = "2010-01-01",
    end_date: str = "2020-12-31",
    prefer : str = 'any',
):
    """
    Generates a test dataframe with pre-defined values.
    These values are meant to be used for testing purposes only.
    The functions generates a standard quantamental dataframe with
    where the value column is populated with pre-defined values.
    These values are simple lines, or waves that are easy to identify
    and differentiate in a plot.
    
    Parameters
    ----------
    :param <List[str]> cids: A list of strings for cids.
    :param <List[str]> xcats: A list of strings for xcats.
    :param <str> start_date: An ISO-formatted date string.
    :param <str> end_date: An ISO-formatted date string.
    :param <str> prefer: A string that specifies the type of line to generate.
        Current choices are: 'linear', 'decreasing-linear', 'sharp-hill',
        'four-bit-sine', 'sine', 'any'.
        See `macrosynergy.management.simulate_quantamental_data.generate_lines`.
    """

    if isinstance(cids, str):
        cids = [cids]
    if isinstance(xcats, str):
        xcats = [xcats]

    dates : pd.DatetimeIndex = pd.bdate_range(start_date, end_date)
    
    df_list: List[pd.DataFrame] = []
    for cid in cids:
        for xcat in xcats:
            df_add: pd.DataFrame = pd.DataFrame(index=dates, 
                                                columns=['real_date','cid',
                                                         'xcat', 'value'])
            df_add['cid'] = cid
            df_add['xcat'] = xcat
            df_add['real_date'] = dates
            df_add['value'] = generate_lines(len(dates), style=prefer)
            df_list.append(df_add)

    return pd.concat(df_list).reset_index(drop=True)

if __name__ == "__main__":

    ser_ar = simulate_ar(100, mean=0, sd_mult=1, ar_coef=0.75)

    cids = ['AUD', 'CAD', 'GBP']
    xcats = ['XR', 'CRY']
    df_cids = pd.DataFrame(index=cids, columns=['earliest', 'latest', 'mean_add',
                                                'sd_mult'])
    df_cids.loc['AUD', ] = ['2010-01-01', '2020-12-31', 0.5, 2]
    df_cids.loc['CAD', ] = ['2011-01-01', '2020-11-30', 0, 1]
    df_cids.loc['GBP', ] = ['2011-01-01', '2020-11-30', -0.2, 0.5]

    df_xcats = pd.DataFrame(index=xcats, columns=['earliest', 'latest',
                                                  'mean_add', 'sd_mult', 'ar_coef',
                                                  'back_coef'])
    df_xcats.loc['XR', ] = ['2010-01-01', '2020-12-31', 0, 1, 0, 0.3]
    df_xcats.loc['CRY', ] = ['2011-01-01', '2020-10-30', 1, 2, 0.9, 0.5]

    dfd = make_qdf(df_cids, df_xcats, back_ar=0.75)



import numpy as np
import pandas as pd
from typing import List, Set
from macrosynergy.management.simulate_quantamental_data import make_qdf
from macrosynergy.management.shape_dfs import reduce_df
import warnings

def _prepare_basket(df: pd.DataFrame, xcat: str, basket: List[str],
                    cids_avl: List[str], complete_cross: bool):
    """
    Categories can be defined over different cross-sections. Will determine the
    respective basket given the available cross-sections for the respective category.

    :param <pd.DataFrame> df: long JPMaQS DataFrame of single category.
    :param <str> xcat: respective category for the relative value calculation.
    :param <pd.DataFrame> basket: cross-sections to be used for the relative value
        benchmark if available.
    :param <List[str] cids_avl: cross-sections available for the category.
    :param <bool> complete_cross: if True, the basket is only calculated if all cross-
        sections, held in the basket, are available for that respective category.
    """

    cids_used: List[str] = sorted(set(basket) & set(cids_avl))
    cids_miss = [b for b in basket if b not in cids_used]

    # Not able to be greater than because of assertion on line 126. If the basket
    # references a cross-section not defined in the DataFrame, an error will be thrown.
    condition = len(cids_used) < len(basket)
    if condition and complete_cross:
        cids_used = []
        err_str: str = (f"The category, {xcat}, is missing {cids_miss} which are included "
              f"in the basket {basket}. Therefore, the category will be excluded "
              f"from the returned DataFrame.")
        print(err_str)
        warnings.warn(err_str, UserWarning)

    elif condition:
        err_str: str = (f"The category, {xcat}, is missing {cids_miss} from the requested "
              f"basket. The new basket will be {cids_used}.")
        print(err_str)
        warnings.warn(err_str, UserWarning)

    # Reduce the DataFrame to the specified basket given the available cross-sections.
    dfb = df[df['cid'].isin(cids_used)]

    return dfb, cids_used

def make_relative_value(df: pd.DataFrame, xcats: List[str], cids: List[str] = None,
                        start: str = None, end: str = None, blacklist: dict = None,
                        basket: List[str] = None, complete_cross: bool = False,
                        rel_meth: str = 'subtract', rel_xcats: List[str] = None,
                        postfix: str = 'R'):
    """
    Returns panel of relative values versus an average of cross-sections.

    :param <pd.DataFrame> df:  standardized JPMaQS DataFrame with the necessary
        columns: 'cid', 'xcat', 'real_date' and 'value'.
    :param <List[str]> xcats: all extended categories for which relative values are to
        be calculated.
    :param <List[str]> cids: cross-sections for which relative values are calculated.
        Default is all cross-section available for the respective category.
    :param <str> start: earliest date in ISO format. Default is None and earliest date
        for which the respective category is available is used.
    :param <str> end: latest date in ISO format. Default is None and latest date for
        which the respective category is available is used.
    :param <dict> blacklist: cross-sections with date ranges that should be excluded from
        the output.
    :param <List[str]> basket: cross-sections to be used for the relative value
        benchmark. The default is every cross-section in the chosen list that is
        available in the DataFrame over the respective time-period.
        However, the basket can be reduced to a valid subset of the available
        cross-sections.
    :param <bool> complete_cross: boolean parameter that outlines whether each category
        is required to have the full set of cross-sections held by the basket parameter
        for a relative value calculation to occur. If set to True, the category will be
        excluded from the output if cross-sections are missing.
        Default is False. If False, the mean, for the relative value, will use the subset
        that is available for that category. For instance, if basket = ['AUD', 'CAD',
        'GBP', 'NZD'] but available cids = ['GBP', 'NZD'], the basket will be implicitly
        updated to basket = ['GBP', 'NZD'] for that respective category.
    :param <str> rel_meth: method for calculating relative value. Default is 'subtract'.
        Alternative is 'divide'.
    :param <List[str]> rel_xcats: extended category name of the relative values. Will
        displace the original category names: xcat + postfix. The order should reflect
        the order of the passed categories.
    :param <str> postfix: acronym to be appended to 'xcat' string to give the name for
        relative value category. Only applies if rel_xcats is None. Default is 'R'

    :return <pd.DataFrame>: standardized DataFrame with the relative values, featuring
        the categories: 'cid', 'xcat', 'real_date' and 'value'.

    """

    expected_columns = ["cid", "xcat", "real_date", "value"]
    col_error = f"The DataFrame must contain the necessary columns: {expected_columns}."
    assert set(expected_columns).issubset(set(df.columns)), col_error

    df = df.loc[:, ["cid", "xcat", "real_date", "value"]]
    df["real_date"] = pd.to_datetime(df["real_date"], format="%Y-%m-%d")

    assert rel_meth in ["subtract", "divide"], "rel_meth must be 'subtract' or 'divide'"

    xcat_error = f"List of categories or single single category string expected. " \
                 f"Received {type(xcats)}."
    assert isinstance(xcats, (list, str)), xcat_error

    if isinstance(xcats, str):
        xcats = [xcats]

    if rel_xcats is not None:
        error_rel_xcat = "List of strings or single string expected for `rel_xcats`."
        assert isinstance(rel_xcats, (list, str)), error_rel_xcat

        if isinstance(rel_xcats, str):
            rel_xcats = [rel_xcats]

        error_length = "`rel_xcats` must have the same number of elements as `xcats`."
        assert len(xcats) == len(rel_xcats), error_length

        rel_xcats_dict = dict(zip(xcats, rel_xcats))

    col_names = ['cid', 'xcat', 'real_date', 'value']
    # Host DataFrame.
    df_out = pd.DataFrame(columns=col_names)

    # Intersect parameter set to False. Therefore, cross-sections across the categories
    # can vary.
    all_cids : List[str] = []
    for cvar in [cids, basket]:
        if cvar is not None:
            all_cids.extend(cvar)
    all_cids = list(set(all_cids))
    if len(all_cids) < 1:
        all_cids = None
    dfx = reduce_df(
        df, xcats, all_cids, start, end, blacklist, out_all=False
    )

    if cids is None:
        # All cross-sections available - union across categories.
        cids = list(dfx['cid'].unique())

    if basket is not None:
        # Basket must be a subset of the available cross-sections.
        miss : Set = set(basket) - set(df['cid'])
        error_basket = f"The basket elements {miss} are not specified or " \
                       f"are not available."
        assert len(miss) == 0, error_basket
    else:
        # Default basket is all available cross-sections.
        basket = cids

    available_xcats = dfx['xcat'].unique()

    if len(cids) == len(basket) == 1:
        run_error = "Computing the relative value on a single cross-section using a " \
                    "basket consisting exclusively of the aforementioned cross-section " \
                    "is an incorrect usage of the function."
        raise RuntimeError(run_error)

    storage = [df_out]
    # Categories can be defined over a different set of cross-sections.
    for i, xcat in enumerate(available_xcats):

        df_xcat = dfx[dfx['xcat'] == xcat]
        available_cids = df_xcat['cid'].unique()

        dfx_xcat = df_xcat[['cid', 'real_date', 'value']]

        dfb, basket = _prepare_basket(
            df=dfx_xcat, xcat=xcat, basket=basket, cids_avl=available_cids,
            complete_cross=complete_cross
        )

        if len(basket) > 1:
            # Mean of (available) cross-sections at each point in time. If all
            # cross-sections defined in the "basket" data structure are not available for
            # a specific date, compute the mean over the available subset.
            bm = dfb.groupby(by='real_date').mean(numeric_only=True)
        elif len(basket) == 1:
            # Relative value is mapped against a single cross-section.
            bm = dfb.set_index('real_date')['value']
        else:
            # Category is not defined over all cross-sections in the basket and
            # 'complete_cross' equals True.
            continue

        dfw = dfx_xcat.pivot(index='real_date', columns='cid', values='value')

        # Computing the relative value is only justified if the number of cross-sections,
        # for the respective date, exceeds one. Therefore, if any rows have only a single
        # cross-section, remove the dates from the DataFrame.
        dfw = dfw[dfw.count(axis=1) > 1]
        # The time-index will be delimited by the respective category.
        dfa = pd.merge(dfw, bm, how='left', left_index=True, right_index=True)

        if rel_meth == 'subtract':
            dfo = dfa[dfw.columns].sub(dfa.loc[:, 'value'], axis=0)
        else:
            dfo = dfa[dfw.columns].div(dfa.loc[:, 'value'], axis=0)

        # Re-stack.
        df_new = dfo.stack().reset_index().rename(
            {'level_1': 'cid', 0: 'value'}, axis=1
        )

        if rel_xcats is None:
            df_new['xcat'] = xcat + postfix
        else:
            df_new['xcat'] = rel_xcats_dict[xcat]

        storage.append(df_new.sort_values(['cid', 'real_date'])[col_names])

    return pd.concat(storage).reset_index(drop=True)


if __name__ == "__main__":

    # Simulate DataFrame.

    cids = ['AUD', 'CAD', 'GBP', 'NZD']
    xcats = ['XR', 'CRY', 'GROWTH', 'INFL']
    df_cids = pd.DataFrame(index=cids, columns=['earliest', 'latest', 'mean_add',
                                                'sd_mult'])
    df_cids.loc['AUD'] = ['2000-01-01', '2020-12-31', 0.1, 1]
    df_cids.loc['CAD'] = ['2001-01-01', '2020-11-30', 0, 1]
    df_cids.loc['GBP'] = ['2002-01-01', '2020-11-30', 0, 2]
    df_cids.loc['NZD'] = ['2002-01-01', '2020-09-30', -0.1, 2]

    df_xcats = pd.DataFrame(index=xcats, columns=['earliest', 'latest', 'mean_add',
                                                  'sd_mult', 'ar_coef', 'back_coef'])
    df_xcats.loc['XR'] = ['2000-01-01', '2020-12-31', 0.1, 1, 0, 0.3]
    df_xcats.loc['CRY'] = ['2000-01-01', '2020-10-30', 1, 2, 0.95, 1]
    df_xcats.loc['GROWTH'] = ['2001-01-01', '2020-10-30', 1, 2, 0.9, 1]
    df_xcats.loc['INFL'] = ['2001-01-01', '2020-10-30', 1, 2, 0.8, 0.5]

    dfd = make_qdf(df_cids, df_xcats, back_ar=0.75)

    # Simulate blacklist
    black = {'AUD': ['2000-01-01', '2003-12-31'], 'GBP': ['2018-01-01', '2100-01-01']}

    # Applications
    dfd_1 = make_relative_value(
        dfd, xcats=["GROWTH", "INFL"], cids=None, blacklist=None, rel_meth='subtract',
        rel_xcats=None, postfix='RV'
    )

    rel_xcats = ["GROWTH_sRV", "INFL_sRV"]
    dfd_1_black = make_relative_value(
        dfd, xcats=["GROWTH", "INFL"], cids=None, blacklist=black, rel_meth='subtract',
        rel_xcats=rel_xcats, postfix='RV'
    )
